"""htsget v1.3 API client."""

from __future__ import annotations

from typing import Any

from ai4gh.core.client import GA4GHClient
from ai4gh.specs.htsget.models import HtsgetClass, HtsgetFormat, HtsgetTicket


class HtsgetClient(GA4GHClient):
    """Client for the GA4GH htsget v1.3 API.

    htsget provides a ticket-based protocol for streaming genomic data.
    It supports reads (BAM/CRAM) and variants (VCF/BCF) endpoints with
    region-based subsetting.
    """

    spec_name = "htsget"
    spec_version = "1.3.0"
    base_path = ""

    async def get_reads(
        self,
        read_id: str,
        *,
        format: HtsgetFormat | None = None,
        data_class: HtsgetClass | None = None,
        reference_name: str | None = None,
        start: int | None = None,
        end: int | None = None,
        fields: list[str] | None = None,
        tags: list[str] | None = None,
        notags: list[str] | None = None,
    ) -> HtsgetTicket:
        """Get a reads ticket for BAM/CRAM data.

        Args:
            read_id: The read identifier.
            format: Desired response format (BAM or CRAM).
            data_class: Request header, body, or both.
            reference_name: Reference sequence name for region subsetting.
            start: Start position (0-based, inclusive).
            end: End position (0-based, exclusive).
            fields: List of BAM fields to include.
            tags: List of tags to include.
            notags: List of tags to exclude.

        Returns:
            An htsget ticket with URLs for downloading the data.
        """
        params = self._build_params(
            format=format,
            data_class=data_class,
            reference_name=reference_name,
            start=start,
            end=end,
            fields=fields,
            tags=tags,
            notags=notags,
        )
        data = await self._request("GET", f"/reads/{read_id}", params=params or None)
        return HtsgetTicket.model_validate(data)

    async def get_variants(
        self,
        variant_id: str,
        *,
        format: HtsgetFormat | None = None,
        data_class: HtsgetClass | None = None,
        reference_name: str | None = None,
        start: int | None = None,
        end: int | None = None,
    ) -> HtsgetTicket:
        """Get a variants ticket for VCF/BCF data.

        Args:
            variant_id: The variant identifier.
            format: Desired response format (VCF or BCF).
            data_class: Request header, body, or both.
            reference_name: Reference sequence name for region subsetting.
            start: Start position (0-based, inclusive).
            end: End position (0-based, exclusive).

        Returns:
            An htsget ticket with URLs for downloading the data.
        """
        params = self._build_params(
            format=format,
            data_class=data_class,
            reference_name=reference_name,
            start=start,
            end=end,
        )
        data = await self._request("GET", f"/variants/{variant_id}", params=params or None)
        return HtsgetTicket.model_validate(data)

    @staticmethod
    def _build_params(
        *,
        format: HtsgetFormat | None = None,
        data_class: HtsgetClass | None = None,
        reference_name: str | None = None,
        start: int | None = None,
        end: int | None = None,
        fields: list[str] | None = None,
        tags: list[str] | None = None,
        notags: list[str] | None = None,
    ) -> dict[str, Any]:
        """Build query parameters for htsget requests."""
        params: dict[str, Any] = {}
        if format is not None:
            params["format"] = format.value
        if data_class is not None:
            params["class"] = data_class.value
        if reference_name is not None:
            params["referenceName"] = reference_name
        if start is not None:
            params["start"] = start
        if end is not None:
            params["end"] = end
        if fields is not None:
            params["fields"] = ",".join(fields)
        if tags is not None:
            params["tags"] = ",".join(tags)
        if notags is not None:
            params["notags"] = ",".join(notags)
        return params



"""Report generation for compliance test results."""

from __future__ import annotations

import json
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ai4gh.compliance.runner import SuiteResult


def to_json(result: SuiteResult) -> str:
    """Generate JSON report from suite result."""
    data = {
        "spec": result.spec,
        "base_url": result.base_url,
        "summary": {
            "total": result.total,
            "passed": result.passed,
            "failed": result.failed,
            "skipped": result.skipped,
            "errors": result.errors,
            "success_rate": round(result.success_rate, 1),
            "duration_ms": round(result.duration_ms, 1),
        },
        "tests": [
            {
                "name": r.name,
                "status": r.status.value,
                "message": r.message,
                "duration_ms": round(r.duration_ms, 1),
                "endpoint": r.endpoint,
            }
            for r in result.results
        ],
    }
    return json.dumps(data, indent=2)


def to_junit_xml(result: SuiteResult) -> str:
    """Generate JUnit XML report from suite result."""
    lines = [
        '<?xml version="1.0" encoding="UTF-8"?>',
        f'<testsuite name="{result.spec}" tests="{result.total}" '
        f'failures="{result.failed}" errors="{result.errors}" '
        f'skipped="{result.skipped}" '
        f'time="{result.duration_ms / 1000:.3f}">',
    ]
    for r in result.results:
        time_s = r.duration_ms / 1000
        lines.append(f'  <testcase name="{r.name}" classname="{result.spec}" time="{time_s:.3f}">')
        if r.status.value == "fail":
            msg = _escape_xml(r.message)
            lines.append(f'    <failure message="{msg}"/>')
        elif r.status.value == "error":
            msg = _escape_xml(r.message)
            lines.append(f'    <error message="{msg}"/>')
        elif r.status.value == "skip":
            msg = _escape_xml(r.message)
            lines.append(f'    <skipped message="{msg}"/>')
        lines.append("  </testcase>")
    lines.append("</testsuite>")
    return "\n".join(lines)


def to_html(result: SuiteResult) -> str:
    """Generate HTML report from suite result."""
    rows = ""
    for r in result.results:
        status_class = {
            "pass": "pass",
            "fail": "fail",
            "skip": "skip",
            "error": "error",
        }.get(r.status.value, "")
        msg = _escape_html(r.message) if r.message else ""
        rows += (
            f"<tr class='{status_class}'>"
            f"<td>{_escape_html(r.name)}</td>"
            f"<td>{r.status.value}</td>"
            f"<td>{msg}</td>"
            f"<td>{r.duration_ms:.1f}ms</td>"
            f"</tr>\n"
        )

    return f"""<!DOCTYPE html>
<html>
<head>
<title>GA4GH Compliance: {_escape_html(result.spec)}</title>
<style>
body {{ font-family: sans-serif; margin: 2em; }}
table {{ border-collapse: collapse; width: 100%; }}
th, td {{ border: 1px solid #ddd; padding: 8px; text-align: left; }}
th {{ background: #f5f5f5; }}
.pass {{ background: #e8f5e9; }}
.fail {{ background: #ffebee; }}
.skip {{ background: #fff3e0; }}
.error {{ background: #fce4ec; }}
.summary {{ margin: 1em 0; }}
</style>
</head>
<body>
<h1>GA4GH Compliance Report: {_escape_html(result.spec)}</h1>
<p>Server: {_escape_html(result.base_url)}</p>
<div class="summary">
<p>Total: {result.total} | Passed: {result.passed} | Failed: {result.failed} |
Skipped: {result.skipped} | Errors: {result.errors} |
Success Rate: {result.success_rate:.1f}% |
Duration: {result.duration_ms:.1f}ms</p>
</div>
<table>
<thead><tr><th>Test</th><th>Status</th><th>Message</th><th>Duration</th></tr></thead>
<tbody>
{rows}</tbody>
</table>
</body>
</html>"""


def _escape_xml(text: str) -> str:
    """Escape special XML characters."""
    return (
        text.replace("&", "&amp;").replace("<", "&lt;").replace(">", "&gt;").replace('"', "&quot;")
    )


def _escape_html(text: str) -> str:
    """Escape special HTML characters."""
    return _escape_xml(text)



"""refget v2 API client."""

from __future__ import annotations

from typing import Any

from ai4gh.core.client import GA4GHClient
from ai4gh.specs.refget.models import RefgetMetadata


class RefgetClient(GA4GHClient):
    """Client for the GA4GH refget v2 API.

    refget provides access to reference sequences by their checksum
    identifiers, supporting subsequence retrieval and metadata lookup.
    """

    spec_name = "refget"
    spec_version = "2.0.0"
    base_path = "/sequence"

    async def get_sequence(
        self,
        checksum: str,
        *,
        start: int | None = None,
        end: int | None = None,
    ) -> str:
        """Retrieve a reference sequence by checksum.

        Args:
            checksum: The sequence checksum (MD5, TRUNC512, or GA4GH).
            start: Start coordinate (0-based, inclusive).
            end: End coordinate (0-based, exclusive).

        Returns:
            The sequence string.
        """
        params: dict[str, Any] = {}
        if start is not None:
            params["start"] = start
        if end is not None:
            params["end"] = end

        headers = {"Accept": "text/plain"}
        url = f"{self.url}/{checksum}"
        auth_headers = await self._get_auth_headers()
        headers.update(auth_headers)

        response = await self._transport.get(url, headers=headers, params=params or None)
        return response.text

    async def get_metadata(self, checksum: str) -> RefgetMetadata:
        """Get metadata for a sequence.

        Args:
            checksum: The sequence checksum.

        Returns:
            Sequence metadata including length and aliases.
        """
        data = await self._request("GET", f"/{checksum}/metadata")
        metadata = data.get("metadata", data)
        return RefgetMetadata.model_validate(metadata)



"""Pydantic models for GA4GH VRS (Variation Representation Specification) v2.

VRS provides a computational framework for the precise representation
and sharing of variation data. This module implements the core value
objects and computed identifier logic.
"""

from __future__ import annotations

import hashlib
import json
from enum import Enum
from typing import Any

from ai4gh.core.types import GA4GHBase


class SequenceString(GA4GHBase):
    """A literal sequence string."""

    sequence: str
    type: str = "SequenceString"


class SequenceReference(GA4GHBase):
    """A reference to a named sequence (by refget accession)."""

    refgetAccession: str
    type: str = "SequenceReference"
    residueAlphabet: str | None = None


class SequenceLocation(GA4GHBase):
    """A contiguous region of a sequence."""

    sequenceReference: SequenceReference | None = None
    start: int | list[int] | None = None
    end: int | list[int] | None = None
    type: str = "SequenceLocation"


class LengthExpression(GA4GHBase):
    """An expression of sequence length."""

    length: int | dict[str, int]
    type: str = "LengthExpression"
    sequence: str | None = None
    repeatSubunitLength: int | None = None


class LiteralSequenceExpression(GA4GHBase):
    """A literal sequence as the state of a variation."""

    sequence: str
    type: str = "LiteralSequenceExpression"


class ReferenceLengthExpression(GA4GHBase):
    """A length expression referencing the reference sequence."""

    length: int | dict[str, int]
    type: str = "ReferenceLengthExpression"
    repeatSubunitLength: int | None = None


class CopyChange(str, Enum):
    """Copy number change types."""

    EFO_0030069 = "efo:0030069"  # complete genomic loss
    EFO_0020073 = "efo:0020073"  # low-level loss
    EFO_0030068 = "efo:0030068"  # loss
    EFO_0030067 = "efo:0030067"  # partial loss
    EFO_0030064 = "efo:0030064"  # regional base ploidy
    EFO_0030070 = "efo:0030070"  # gain
    EFO_0030071 = "efo:0030071"  # low-level gain
    EFO_0030072 = "efo:0030072"  # high-level gain


class Allele(GA4GHBase):
    """VRS Allele — a contiguous change at a SequenceLocation."""

    location: SequenceLocation
    state: LiteralSequenceExpression | ReferenceLengthExpression
    type: str = "Allele"
    id: str | None = None


class Haplotype(GA4GHBase):
    """VRS Haplotype — a set of co-occurring Alleles on a single molecule."""

    members: list[Allele | dict[str, Any]]
    type: str = "Haplotype"
    id: str | None = None


class CopyNumberCount(GA4GHBase):
    """VRS CopyNumberCount — an absolute copy number at a location."""

    location: SequenceLocation
    copies: int | dict[str, int]
    type: str = "CopyNumberCount"
    id: str | None = None


class CopyNumberChange(GA4GHBase):
    """VRS CopyNumberChange — a relative copy number change at a location."""

    location: SequenceLocation
    copy_change: CopyChange
    type: str = "CopyNumberChange"
    id: str | None = None


class Adjacency(GA4GHBase):
    """VRS Adjacency — a junction between two sequence locations."""

    adj_left: SequenceLocation
    adj_right: SequenceLocation
    type: str = "Adjacency"
    id: str | None = None
    linker_sequence: str | None = None


def ga4gh_digest(data: dict[str, Any]) -> str:
    """Compute a GA4GH digest for a VRS object.

    The GA4GH digest uses SHA-512 truncated to 24 bytes, then
    base64url-encoded (no padding).

    Args:
        data: A serializable VRS value object dict.

    Returns:
        The base64url-encoded truncated SHA-512 digest.
    """
    import base64

    canonical = json.dumps(data, sort_keys=True, separators=(",", ":"))
    sha = hashlib.sha512(canonical.encode("utf-8")).digest()[:24]
    return base64.urlsafe_b64encode(sha).rstrip(b"=").decode("ascii")


def ga4gh_identify(vrs_obj: dict[str, Any], vrs_type: str) -> str:
    """Compute a GA4GH Computed Identifier for a VRS object.

    The identifier has the form: ga4gh:{type}.{digest}

    Args:
        vrs_obj: A serializable VRS value object dict.
        vrs_type: The VRS type prefix (e.g., "VA" for Allele, "VH" for Haplotype).

    Returns:
        A GA4GH computed identifier string.
    """
    digest = ga4gh_digest(vrs_obj)
    return f"ga4gh:{vrs_type}.{digest}"



"""Pydantic models for GA4GH VA-Spec (Variant Annotation Specification).

VA-Spec provides a framework for representing variant annotations,
evidence, and interpretations. It defines structures for clinical
assertions about variants.
"""

from __future__ import annotations

from enum import Enum

from ai4gh.core.types import GA4GHBase


class Direction(str, Enum):
    """Direction of evidence support."""

    SUPPORTS = "supports"
    DISPUTES = "disputes"
    UNCERTAIN = "uncertain"


class Classification(str, Enum):
    """Variant classification categories (ACMG/AMP)."""

    PATHOGENIC = "pathogenic"
    LIKELY_PATHOGENIC = "likely_pathogenic"
    UNCERTAIN_SIGNIFICANCE = "uncertain_significance"
    LIKELY_BENIGN = "likely_benign"
    BENIGN = "benign"


class EvidenceLevel(str, Enum):
    """Strength of evidence."""

    VERY_STRONG = "very_strong"
    STRONG = "strong"
    MODERATE = "moderate"
    SUPPORTING = "supporting"
    STANDALONE = "standalone"


class Agent(GA4GHBase):
    """An agent (person, organization, or software) making contributions."""

    id: str | None = None
    type: str = "Agent"
    label: str | None = None
    description: str | None = None


class Contribution(GA4GHBase):
    """A contribution by an agent to an annotation."""

    type: str = "Contribution"
    contributor: Agent | None = None
    date: str | None = None
    activity: str | None = None


class Method(GA4GHBase):
    """A method used to generate evidence or interpretations."""

    id: str | None = None
    type: str = "Method"
    label: str | None = None
    url: str | None = None
    version: str | None = None


class Document(GA4GHBase):
    """A reference document (publication, guideline, etc.)."""

    id: str | None = None
    type: str = "Document"
    label: str | None = None
    url: str | None = None
    pmid: str | None = None
    doi: str | None = None


class EvidenceLine(GA4GHBase):
    """A line of evidence supporting or disputing an assertion."""

    id: str | None = None
    type: str = "EvidenceLine"
    direction: Direction | None = None
    strength: EvidenceLevel | None = None
    description: str | None = None
    evidence_items: list[Document] | None = None
    contributions: list[Contribution] | None = None


class Statement(GA4GHBase):
    """A statement (assertion) about a variant.

    This is the core unit of variant annotation: a claim
    about a variant's clinical significance, supported by
    evidence lines.
    """

    id: str | None = None
    type: str = "Statement"
    subject: str | None = None
    predicate: str | None = None
    object_descriptor: str | None = None
    classification: Classification | None = None
    direction: Direction | None = None
    description: str | None = None
    evidence_lines: list[EvidenceLine] | None = None
    method: Method | None = None
    contributions: list[Contribution] | None = None
    specified_by: Method | None = None



"""Beacon v2 API client."""

from __future__ import annotations

from typing import Any

from ai4gh.core.client import GA4GHClient
from ai4gh.specs.beacon.models import (
    BeaconInfo,
    BeaconQuery,
    BeaconRequest,
    BeaconRequestMeta,
    BeaconResponse,
    Granularity,
)


class BeaconClient(GA4GHClient):
    """Client for the GA4GH Beacon v2 API.

    Beacon v2 provides a discovery protocol for genomic and clinical data.
    It supports queries across multiple entry types including individuals,
    biosamples, genomic variants, cohorts, runs, and analyses.
    """

    spec_name = "beacon"
    spec_version = "2.0.0"
    base_path = ""

    async def get_info(self) -> BeaconInfo:
        """Get Beacon info (/ endpoint).

        Returns:
            Beacon informational response.
        """
        data = await self._request("GET", "/")
        response = data.get("response", data)
        return BeaconInfo.model_validate(response)

    async def get_configuration(self) -> dict[str, Any]:
        """Get Beacon configuration.

        Returns:
            Beacon configuration object.
        """
        data: dict[str, Any] = await self._request("GET", "/configuration")
        return data

    async def get_map(self) -> dict[str, Any]:
        """Get Beacon map (available entry types and endpoints).

        Returns:
            Beacon map object.
        """
        data: dict[str, Any] = await self._request("GET", "/map")
        return data

    async def get_entry_types(self) -> dict[str, Any]:
        """Get available entry types.

        Returns:
            Entry types configuration.
        """
        data: dict[str, Any] = await self._request("GET", "/entry_types")
        return data

    async def get_filtering_terms(
        self,
        *,
        skip: int | None = None,
        limit: int | None = None,
    ) -> dict[str, Any]:
        """Get available filtering terms.

        Args:
            skip: Number of results to skip.
            limit: Maximum number of results.

        Returns:
            Filtering terms response.
        """
        params: dict[str, Any] = {}
        if skip is not None:
            params["skip"] = skip
        if limit is not None:
            params["limit"] = limit
        result: dict[str, Any] = await self._request(
            "GET", "/filtering_terms", params=params or None
        )
        return result

    async def query(
        self,
        entry_type: str,
        *,
        filters: list[dict[str, Any]] | None = None,
        request_parameters: dict[str, Any] | None = None,
        granularity: Granularity | None = None,
        skip: int | None = None,
        limit: int | None = None,
        test_mode: bool = False,
    ) -> BeaconResponse:
        """Query a Beacon entry type (POST).

        Args:
            entry_type: Entry type to query (e.g., "individuals", "biosamples",
                "g_variants", "cohorts", "runs", "analyses").
            filters: List of filter objects.
            request_parameters: Additional request parameters.
            granularity: Response granularity (boolean, count, record).
            skip: Number of results to skip.
            limit: Maximum number of results.
            test_mode: If true, return test data.

        Returns:
            Beacon response.
        """
        query = BeaconQuery(
            filters=None,
            request_parameters=request_parameters,
            test_mode=test_mode,
        )
        if skip is not None or limit is not None:
            query.pagination = {}
            if skip is not None:
                query.pagination["skip"] = skip
            if limit is not None:
                query.pagination["limit"] = limit

        meta = BeaconRequestMeta()
        if granularity:
            meta.requested_schemas = []

        request = BeaconRequest(meta=meta, query=query)
        body = request.model_dump(exclude_none=True)

        if filters:
            body.setdefault("query", {})["filters"] = filters

        if granularity:
            body.setdefault("query", {})["requestedGranularity"] = granularity.value

        data = await self._request("POST", f"/{entry_type}", json=body)
        return BeaconResponse.model_validate(data)

    async def query_get(
        self,
        entry_type: str,
        *,
        skip: int | None = None,
        limit: int | None = None,
    ) -> BeaconResponse:
        """Query a Beacon entry type (GET).

        Args:
            entry_type: Entry type to query.
            skip: Number of results to skip.
            limit: Maximum number of results.

        Returns:
            Beacon response.
        """
        params: dict[str, Any] = {}
        if skip is not None:
            params["skip"] = skip
        if limit is not None:
            params["limit"] = limit
        data = await self._request("GET", f"/{entry_type}", params=params or None)
        return BeaconResponse.model_validate(data)



"""Tests for CLI compliance commands."""

from __future__ import annotations

from typer.testing import CliRunner

from ai4gh.cli.main import app

runner = CliRunner()


class TestListSpecs:
    def test_list_specs(self) -> None:
        result = runner.invoke(app, ["list-specs"])
        assert result.exit_code == 0
        assert "beacon" in result.output
        assert "drs" in result.output

    def test_list_specs_all_seven(self) -> None:
        result = runner.invoke(app, ["list-specs"])
        for spec in ("beacon", "drs", "wes", "tes", "trs", "htsget", "refget"):
            assert spec in result.output


class TestTestCommand:
    def test_unknown_spec(self) -> None:
        result = runner.invoke(app, ["test", "nonexistent", "http://example.com"])
        assert result.exit_code == 1

    def test_help(self) -> None:
        result = runner.invoke(app, ["test", "--help"])
        assert result.exit_code == 0
        assert "spec" in result.output.lower()
        assert "url" in result.output.lower()


class TestConfigCommands:
    def test_config_show(self) -> None:
        result = runner.invoke(app, ["config", "show"])
        assert result.exit_code == 0
        assert "Timeout" in result.output

    def test_config_help(self) -> None:
        result = runner.invoke(app, ["config", "--help"])
        assert result.exit_code == 0
        assert "show" in result.output
        assert "init" in result.output



"""Tests for WES models."""

from __future__ import annotations

from ai4gh.specs.wes.models import (
    RunDescription,
    RunId,
    RunListResponse,
    RunLog,
    RunRequest,
    RunStatus,
    ServiceInfo,
    State,
)


class TestState:
    def test_enum_values(self) -> None:
        assert State.UNKNOWN == "UNKNOWN"
        assert State.QUEUED == "QUEUED"
        assert State.RUNNING == "RUNNING"
        assert State.COMPLETE == "COMPLETE"
        assert State.EXECUTOR_ERROR == "EXECUTOR_ERROR"
        assert State.SYSTEM_ERROR == "SYSTEM_ERROR"
        assert State.CANCELED == "CANCELED"
        assert State.CANCELING == "CANCELING"

    def test_from_string(self) -> None:
        assert State("RUNNING") is State.RUNNING


class TestRunLog:
    def test_minimal(self) -> None:
        log = RunLog()
        assert log.name is None
        assert log.cmd is None
        assert log.exit_code is None

    def test_full(self) -> None:
        log = RunLog(
            name="step1",
            cmd=["echo", "hello"],
            start_time="2024-01-01T00:00:00Z",
            end_time="2024-01-01T00:01:00Z",
            stdout="/logs/stdout",
            stderr="/logs/stderr",
            exit_code=0,
        )
        assert log.name == "step1"
        assert log.cmd == ["echo", "hello"]
        assert log.exit_code == 0


class TestRunRequest:
    def test_minimal(self) -> None:
        req = RunRequest()
        assert req.workflow_url is None
        assert req.workflow_type is None

    def test_full(self) -> None:
        req = RunRequest(
            workflow_url="https://example.com/workflow.cwl",
            workflow_type="CWL",
            workflow_type_version="v1.0",
            workflow_params={"input": "data.bam"},
            tags={"project": "test"},
        )
        assert req.workflow_url == "https://example.com/workflow.cwl"
        assert req.workflow_type == "CWL"
        assert req.tags == {"project": "test"}


class TestRunStatus:
    def test_creation(self) -> None:
        status = RunStatus(run_id="run-1", state=State.RUNNING)
        assert status.run_id == "run-1"
        assert status.state is State.RUNNING

    def test_from_dict(self) -> None:
        status = RunStatus.model_validate({"run_id": "run-2", "state": "COMPLETE"})
        assert status.state is State.COMPLETE


class TestRunListResponse:
    def test_empty(self) -> None:
        resp = RunListResponse()
        assert resp.runs is None
        assert resp.next_page_token is None

    def test_with_runs(self) -> None:
        resp = RunListResponse(
            runs=[RunStatus(run_id="r1", state=State.RUNNING)],
            next_page_token="token-2",
        )
        assert len(resp.runs) == 1
        assert resp.next_page_token == "token-2"


class TestRunDescription:
    def test_minimal(self) -> None:
        desc = RunDescription(run_id="run-1")
        assert desc.run_id == "run-1"
        assert desc.state is None
        assert desc.run_log is None
        assert desc.task_logs is None

    def test_full(self) -> None:
        desc = RunDescription(
            run_id="run-1",
            state=State.COMPLETE,
            request=RunRequest(workflow_url="https://example.com/wf.cwl"),
            run_log=RunLog(name="main", exit_code=0),
            task_logs=[RunLog(name="task1", exit_code=0)],
            outputs={"output1": "s3://bucket/result.bam"},
        )
        assert desc.state is State.COMPLETE
        assert desc.run_log.exit_code == 0
        assert len(desc.task_logs) == 1
        assert "output1" in desc.outputs


class TestRunId:
    def test_creation(self) -> None:
        rid = RunId(run_id="run-123")
        assert rid.run_id == "run-123"


class TestServiceInfo:
    def test_minimal(self) -> None:
        info = ServiceInfo()
        assert info.supported_wes_versions is None

    def test_full(self) -> None:
        info = ServiceInfo(
            supported_wes_versions=["1.0.0", "1.1.0"],
            supported_filesystem_protocols=["s3", "gs"],
            auth_instructions_url="https://example.com/auth",
        )
        assert "1.1.0" in info.supported_wes_versions
        assert "s3" in info.supported_filesystem_protocols


class TestExtraFields:
    def test_allows_extra(self) -> None:
        log = RunLog.model_validate({"name": "step1", "custom_field": "custom_value"})
        assert log.name == "step1"
        assert log.custom_field == "custom_value"



"""Tests for DUO matcher."""

from __future__ import annotations

from ai4gh.specs.duo.matcher import DUOMatcher
from ai4gh.specs.duo.models import (
    DataUseCondition,
    DUOModifier,
    ResearchPurpose,
)


class TestDUOMatcher:
    def setup_method(self) -> None:
        self.matcher = DUOMatcher()

    def test_no_conditions(self) -> None:
        purpose = ResearchPurpose(codes=["DUO:0000042"])
        result = self.matcher.check_compatibility([], purpose)
        assert result.compatible is True
        assert "No data use conditions" in result.reason

    def test_permissive_no_restriction(self) -> None:
        conditions = [DataUseCondition(code=DUOModifier.NO_RESTRICTION.value)]
        purpose = ResearchPurpose(codes=["DUO:0000042"])
        result = self.matcher.check_compatibility(conditions, purpose)
        assert result.compatible is True

    def test_permissive_general_research(self) -> None:
        conditions = [DataUseCondition(code=DUOModifier.GENERAL_RESEARCH_USE.value)]
        purpose = ResearchPurpose(codes=["DUO:0000042"])
        result = self.matcher.check_compatibility(conditions, purpose)
        assert result.compatible is True

    def test_disease_specific_match(self) -> None:
        conditions = [
            DataUseCondition(
                code=DUOModifier.DISEASE_SPECIFIC.value,
                disease_id="MONDO:0005015",
            )
        ]
        purpose = ResearchPurpose(
            codes=["DUO:0000007"],
            disease_id="MONDO:0005015",
        )
        result = self.matcher.check_compatibility(conditions, purpose)
        assert result.compatible is True

    def test_disease_specific_mismatch(self) -> None:
        conditions = [
            DataUseCondition(
                code=DUOModifier.DISEASE_SPECIFIC.value,
                disease_id="MONDO:0005015",
            )
        ]
        purpose = ResearchPurpose(
            codes=["DUO:0000007"],
            disease_id="MONDO:0009999",
        )
        result = self.matcher.check_compatibility(conditions, purpose)
        assert result.compatible is False
        assert len(result.blocking_conditions) == 1

    def test_not_for_profit_academic(self) -> None:
        conditions = [DataUseCondition(code=DUOModifier.NOT_FOR_PROFIT.value)]
        purpose = ResearchPurpose(codes=["DUO:0000042"], is_for_profit=False)
        result = self.matcher.check_compatibility(conditions, purpose)
        assert result.compatible is True

    def test_not_for_profit_commercial(self) -> None:
        conditions = [DataUseCondition(code=DUOModifier.NOT_FOR_PROFIT.value)]
        purpose = ResearchPurpose(codes=["DUO:0000042"], is_for_profit=True)
        result = self.matcher.check_compatibility(conditions, purpose)
        assert result.compatible is False

    def test_non_commercial_commercial(self) -> None:
        conditions = [DataUseCondition(code=DUOModifier.NON_COMMERCIAL.value)]
        purpose = ResearchPurpose(codes=["DUO:0000042"], is_for_profit=True)
        result = self.matcher.check_compatibility(conditions, purpose)
        assert result.compatible is False

    def test_institution_specific_match(self) -> None:
        conditions = [
            DataUseCondition(
                code=DUOModifier.INSTITUTION_SPECIFIC.value,
                institution="MIT",
            )
        ]
        purpose = ResearchPurpose(codes=["DUO:0000042"], institution="MIT")
        result = self.matcher.check_compatibility(conditions, purpose)
        assert result.compatible is True

    def test_institution_specific_mismatch(self) -> None:
        conditions = [
            DataUseCondition(
                code=DUOModifier.INSTITUTION_SPECIFIC.value,
                institution="MIT",
            )
        ]
        purpose = ResearchPurpose(codes=["DUO:0000042"], institution="Stanford")
        result = self.matcher.check_compatibility(conditions, purpose)
        assert result.compatible is False

    def test_project_specific_match(self) -> None:
        conditions = [
            DataUseCondition(
                code=DUOModifier.PROJECT_SPECIFIC.value,
                project="GENOMICS-2024",
            )
        ]
        purpose = ResearchPurpose(codes=["DUO:0000042"], project="GENOMICS-2024")
        result = self.matcher.check_compatibility(conditions, purpose)
        assert result.compatible is True

    def test_project_specific_mismatch(self) -> None:
        conditions = [
            DataUseCondition(
                code=DUOModifier.PROJECT_SPECIFIC.value,
                project="GENOMICS-2024",
            )
        ]
        purpose = ResearchPurpose(codes=["DUO:0000042"], project="OTHER-PROJECT")
        result = self.matcher.check_compatibility(conditions, purpose)
        assert result.compatible is False

    def test_multiple_conditions_all_pass(self) -> None:
        conditions = [
            DataUseCondition(code=DUOModifier.GENERAL_RESEARCH_USE.value),
            DataUseCondition(code=DUOModifier.NOT_FOR_PROFIT.value),
        ]
        purpose = ResearchPurpose(codes=["DUO:0000042"], is_for_profit=False)
        result = self.matcher.check_compatibility(conditions, purpose)
        assert result.compatible is True
        assert len(result.matched_conditions) == 2

    def test_multiple_conditions_one_blocks(self) -> None:
        conditions = [
            DataUseCondition(code=DUOModifier.GENERAL_RESEARCH_USE.value),
            DataUseCondition(code=DUOModifier.NOT_FOR_PROFIT.value),
        ]
        purpose = ResearchPurpose(codes=["DUO:0000042"], is_for_profit=True)
        result = self.matcher.check_compatibility(conditions, purpose)
        assert result.compatible is False
        assert len(result.blocking_conditions) == 1
        assert len(result.matched_conditions) == 1

    def test_unknown_condition_passes(self) -> None:
        conditions = [DataUseCondition(code="DUO:9999999")]
        purpose = ResearchPurpose(codes=["DUO:0000042"])
        result = self.matcher.check_compatibility(conditions, purpose)
        assert result.compatible is True



"""Pydantic models for GA4GH Data Use Ontology (DUO).

DUO provides a standard vocabulary for describing data use conditions
and restrictions. This module models DUO terms and provides utilities
for checking compatibility between dataset conditions and researcher purposes.
"""

from __future__ import annotations

from enum import Enum

from ai4gh.core.types import GA4GHBase


class DUOModifier(str, Enum):
    """Common DUO modifier terms."""

    NO_RESTRICTION = "DUO:0000004"
    GENERAL_RESEARCH_USE = "DUO:0000042"
    HEALTH_MEDICAL_BIOMEDICAL = "DUO:0000006"
    DISEASE_SPECIFIC = "DUO:0000007"
    NOT_FOR_PROFIT = "DUO:0000018"
    NON_COMMERCIAL = "DUO:0000046"
    PUBLICATION_REQUIRED = "DUO:0000019"
    COLLABORATION_REQUIRED = "DUO:0000020"
    ETHICS_APPROVAL_REQUIRED = "DUO:0000021"
    GEOGRAPHICAL_RESTRICTION = "DUO:0000022"
    INSTITUTION_SPECIFIC = "DUO:0000028"
    PROJECT_SPECIFIC = "DUO:0000027"
    TIME_LIMIT = "DUO:0000025"
    USER_SPECIFIC = "DUO:0000026"
    RETURN_TO_DATABASE = "DUO:0000029"
    NO_GENERAL_METHODS = "DUO:0000015"
    GENETIC_STUDIES_ONLY = "DUO:0000016"
    POPULATION_ORIGINS = "DUO:0000011"


class DUOTerm(GA4GHBase):
    """A DUO term with optional modifiers."""

    id: str
    label: str | None = None
    description: str | None = None
    shorthand: str | None = None
    modifiers: list[DUOTerm] | None = None


class DataUseCondition(GA4GHBase):
    """A data use condition applied to a dataset."""

    code: str
    label: str | None = None
    description: str | None = None
    disease_id: str | None = None
    geographic_restriction: str | None = None
    institution: str | None = None
    project: str | None = None
    expiration_date: str | None = None


class ResearchPurpose(GA4GHBase):
    """A researcher's declared purpose for data access."""

    codes: list[str]
    is_for_profit: bool = False
    disease_id: str | None = None
    institution: str | None = None
    project: str | None = None


class CompatibilityResult(GA4GHBase):
    """Result of checking research purpose against data use conditions."""

    compatible: bool
    reason: str
    matched_conditions: list[DataUseCondition] | None = None
    blocking_conditions: list[DataUseCondition] | None = None



"""GA4GH Data Use Ontology (DUO) plugin."""

from __future__ import annotations

from ai4gh.specs.duo.matcher import DUOMatcher
from ai4gh.specs.duo.models import (
    CompatibilityResult,
    DataUseCondition,
    DUOModifier,
    DUOTerm,
    ResearchPurpose,
)

__all__ = [
    "CompatibilityResult",
    "DUOMatcher",
    "DUOModifier",
    "DUOTerm",
    "DataUseCondition",
    "ResearchPurpose",
]



"""TRS v2.0 compliance test suite."""

from __future__ import annotations

import httpx

from ai4gh.compliance.runner import ComplianceSuite, TestResult, TestStatus, register_suite


@register_suite("trs")
class TRSComplianceSuite(ComplianceSuite):
    """Test GA4GH TRS API compliance."""

    spec_name = "trs"
    spec_version = "2.0.1"

    async def test_service_info(self) -> TestResult:
        """GET /ga4gh/trs/v2/service-info returns service info."""
        async with httpx.AsyncClient() as client:
            resp = await client.get(f"{self.base_url}/ga4gh/trs/v2/service-info")
        if resp.status_code != 200:
            return TestResult(
                name="test_service_info",
                status=TestStatus.FAIL,
                message=f"Expected 200, got {resp.status_code}",
                spec=self.spec_name,
                endpoint="/ga4gh/trs/v2/service-info",
            )
        return TestResult(
            name="test_service_info",
            status=TestStatus.PASS,
            spec=self.spec_name,
            endpoint="/ga4gh/trs/v2/service-info",
        )

    async def test_tool_classes(self) -> TestResult:
        """GET /ga4gh/trs/v2/toolClasses returns tool classes."""
        async with httpx.AsyncClient() as client:
            resp = await client.get(f"{self.base_url}/ga4gh/trs/v2/toolClasses")
        if resp.status_code != 200:
            return TestResult(
                name="test_tool_classes",
                status=TestStatus.FAIL,
                message=f"Expected 200, got {resp.status_code}",
                spec=self.spec_name,
                endpoint="/ga4gh/trs/v2/toolClasses",
            )
        data = resp.json()
        if not isinstance(data, list):
            return TestResult(
                name="test_tool_classes",
                status=TestStatus.FAIL,
                message="Expected array response",
                spec=self.spec_name,
                endpoint="/ga4gh/trs/v2/toolClasses",
            )
        return TestResult(
            name="test_tool_classes",
            status=TestStatus.PASS,
            spec=self.spec_name,
            endpoint="/ga4gh/trs/v2/toolClasses",
        )

    async def test_tools_endpoint(self) -> TestResult:
        """GET /ga4gh/trs/v2/tools returns tools list."""
        async with httpx.AsyncClient() as client:
            resp = await client.get(f"{self.base_url}/ga4gh/trs/v2/tools")
        if resp.status_code not in (200, 401, 403):
            return TestResult(
                name="test_tools_endpoint",
                status=TestStatus.FAIL,
                message=f"Expected 200/401/403, got {resp.status_code}",
                spec=self.spec_name,
                endpoint="/ga4gh/trs/v2/tools",
            )
        return TestResult(
            name="test_tools_endpoint",
            status=TestStatus.PASS,
            spec=self.spec_name,
            endpoint="/ga4gh/trs/v2/tools",
        )



"""GA4GH WES (Workflow Execution Service) v1.1 plugin."""

from __future__ import annotations

from ai4gh.specs.wes.client import WESClient
from ai4gh.specs.wes.models import (
    RunDescription,
    RunId,
    RunListResponse,
    RunLog,
    RunRequest,
    RunStatus,
    State,
)

__all__ = [
    "RunDescription",
    "RunId",
    "RunListResponse",
    "RunLog",
    "RunRequest",
    "RunStatus",
    "State",
    "WESClient",
]



"""Tests for refget models."""

from __future__ import annotations

from ai4gh.specs.refget.models import RefgetAlias, RefgetMetadata


class TestRefgetAlias:
    def test_basic(self) -> None:
        alias = RefgetAlias(alias="chr1", naming_authority="UCSC")
        assert alias.alias == "chr1"
        assert alias.naming_authority == "UCSC"


class TestRefgetMetadata:
    def test_minimal(self) -> None:
        m = RefgetMetadata(id="seq-1", length=1000)
        assert m.id == "seq-1"
        assert m.length == 1000
        assert m.md5 is None
        assert m.aliases is None

    def test_full(self) -> None:
        m = RefgetMetadata(
            id="seq-1",
            length=248956422,
            md5="6aef897c3d6ff0c78aff06ac189178dd",
            trunc512="f7f5e8da3ec02a5f7c3e2f3fcf53f8a2",
            ga4gh="SQ.abc123",
            aliases=[
                RefgetAlias(alias="chr1", naming_authority="UCSC"),
                RefgetAlias(alias="1", naming_authority="ensembl"),
            ],
        )
        assert m.md5 is not None
        assert m.ga4gh == "SQ.abc123"
        assert m.aliases is not None
        assert len(m.aliases) == 2
        assert m.aliases[0].alias == "chr1"



"""Tests for MCP prompts."""

from __future__ import annotations

from ai4gh.mcp.prompts import get_prompt, list_prompts


class TestListPrompts:
    def test_returns_list(self) -> None:
        prompts = list_prompts()
        assert isinstance(prompts, list)

    def test_has_prompts(self) -> None:
        prompts = list_prompts()
        assert len(prompts) == 5

    def test_prompt_structure(self) -> None:
        prompts = list_prompts()
        for p in prompts:
            assert "name" in p
            assert "description" in p
            assert "arguments" in p

    def test_prompt_names(self) -> None:
        prompts = list_prompts()
        names = {p["name"] for p in prompts}
        assert names == {
            "find_variants",
            "patient_summary",
            "rare_disease_search",
            "workflow_submit",
            "compliance_report",
        }


class TestGetPrompt:
    def test_find_variants(self) -> None:
        result = get_prompt("find_variants", {"gene": "TP53"})
        assert "TP53" in result.messages[0].content.text
        assert result.description

    def test_find_variants_defaults(self) -> None:
        result = get_prompt("find_variants", {})
        assert "BRCA1" in result.messages[0].content.text

    def test_patient_summary(self) -> None:
        result = get_prompt("patient_summary", {"phenopacket_json": '{"id": "test"}'})
        assert '{"id": "test"}' in result.messages[0].content.text

    def test_rare_disease_search(self) -> None:
        result = get_prompt("rare_disease_search", {"disease": "OMIM:100100"})
        assert "OMIM:100100" in result.messages[0].content.text

    def test_rare_disease_search_with_phenotypes(self) -> None:
        result = get_prompt(
            "rare_disease_search",
            {"disease": "test", "phenotypes": "HP:0001234,HP:0005678"},
        )
        assert "HP:0001234" in result.messages[0].content.text

    def test_workflow_submit(self) -> None:
        result = get_prompt(
            "workflow_submit",
            {"workflow_url": "https://example.com/wf.cwl", "server_url": "https://wes.example.com"},
        )
        assert "https://example.com/wf.cwl" in result.messages[0].content.text

    def test_compliance_report(self) -> None:
        result = get_prompt(
            "compliance_report",
            {"server_url": "https://beacon.example.com", "spec": "beacon"},
        )
        assert "beacon" in result.messages[0].content.text

    def test_unknown_prompt(self) -> None:
        result = get_prompt("nonexistent", {})
        assert "Unknown prompt" in result.messages[0].content.text

    def test_none_arguments(self) -> None:
        result = get_prompt("find_variants", None)
        assert result.messages



"""Tests for htsget client."""

from __future__ import annotations

import pytest
from pytest_httpx import HTTPXMock

from ai4gh.specs.htsget.client import HtsgetClient
from ai4gh.specs.htsget.models import HtsgetClass, HtsgetFormat

READS_TICKET_RESPONSE = {
    "htsget": {
        "format": "BAM",
        "urls": [
            {
                "url": "https://example.com/header",
                "headers": {"Authorization": "Bearer x"},
                "data_class": "header",
            },
            {
                "url": "https://example.com/body",
                "data_class": "body",
            },
        ],
    }
}

VARIANTS_TICKET_RESPONSE = {
    "htsget": {
        "format": "VCF",
        "urls": [
            {"url": "https://example.com/variants/block1"},
        ],
    }
}


@pytest.fixture
def htsget_client() -> HtsgetClient:
    return HtsgetClient(base_url="https://htsget.example.com")


class TestHtsgetClient:
    def test_spec_attrs(self) -> None:
        client = HtsgetClient(base_url="https://example.com")
        assert client.spec_name == "htsget"
        assert client.spec_version == "1.3.0"

    async def test_get_reads(self, htsget_client: HtsgetClient, httpx_mock: HTTPXMock) -> None:
        httpx_mock.add_response(json=READS_TICKET_RESPONSE)
        ticket = await htsget_client.get_reads("sample1")
        assert ticket.htsget.format == HtsgetFormat.BAM
        assert len(ticket.htsget.urls) == 2

    async def test_get_reads_with_params(
        self, htsget_client: HtsgetClient, httpx_mock: HTTPXMock
    ) -> None:
        httpx_mock.add_response(json=READS_TICKET_RESPONSE)
        ticket = await htsget_client.get_reads(
            "sample1",
            format=HtsgetFormat.CRAM,
            data_class=HtsgetClass.header,
            reference_name="chr1",
            start=100,
            end=200,
            fields=["QNAME", "FLAG"],
            tags=["RG"],
            notags=["OQ"],
        )
        assert ticket.htsget.format == HtsgetFormat.BAM
        request = httpx_mock.get_requests()[0]
        url_str = str(request.url)
        assert "format=CRAM" in url_str
        assert "class=header" in url_str
        assert "referenceName=chr1" in url_str
        assert "start=100" in url_str
        assert "end=200" in url_str

    async def test_get_variants(self, htsget_client: HtsgetClient, httpx_mock: HTTPXMock) -> None:
        httpx_mock.add_response(json=VARIANTS_TICKET_RESPONSE)
        ticket = await htsget_client.get_variants("variant1")
        assert ticket.htsget.format == HtsgetFormat.VCF
        assert len(ticket.htsget.urls) == 1

    async def test_get_variants_with_region(
        self, htsget_client: HtsgetClient, httpx_mock: HTTPXMock
    ) -> None:
        httpx_mock.add_response(json=VARIANTS_TICKET_RESPONSE)
        ticket = await htsget_client.get_variants(
            "variant1",
            format=HtsgetFormat.BCF,
            reference_name="chr22",
            start=1000,
            end=5000,
        )
        assert ticket.htsget.format == HtsgetFormat.VCF
        request = httpx_mock.get_requests()[0]
        url_str = str(request.url)
        assert "format=BCF" in url_str
        assert "referenceName=chr22" in url_str

    def test_build_params_empty(self) -> None:
        params = HtsgetClient._build_params()
        assert params == {}

    def test_build_params_full(self) -> None:
        params = HtsgetClient._build_params(
            format=HtsgetFormat.BAM,
            data_class=HtsgetClass.body,
            reference_name="chr1",
            start=0,
            end=1000,
            fields=["QNAME"],
            tags=["RG", "BC"],
            notags=["OQ"],
        )
        assert params["format"] == "BAM"
        assert params["class"] == "body"
        assert params["referenceName"] == "chr1"
        assert params["start"] == 0
        assert params["end"] == 1000
        assert params["fields"] == "QNAME"
        assert params["tags"] == "RG,BC"
        assert params["notags"] == "OQ"



"""GA4GH VA-Spec (Variant Annotation Specification) plugin."""

from __future__ import annotations

from ai4gh.specs.va_spec.models import (
    Agent,
    Classification,
    Contribution,
    Direction,
    Document,
    EvidenceLevel,
    EvidenceLine,
    Method,
    Statement,
)

__all__ = [
    "Agent",
    "Classification",
    "Contribution",
    "Direction",
    "Document",
    "EvidenceLevel",
    "EvidenceLine",
    "Method",
    "Statement",
]



"""Pydantic models for GA4GH Cat-VRS (Categorical Variation Representation).

Cat-VRS extends VRS with categorical variation concepts:
constraints on variation that describe classes of variation
rather than specific instances (e.g., "any loss-of-function
variant in BRCA1").
"""

from __future__ import annotations

from enum import Enum

from ai4gh.core.types import GA4GHBase
from ai4gh.specs.vrs.models import (  # noqa: TC001 — needed at runtime by Pydantic
    CopyChange,
    SequenceLocation,
    SequenceReference,
)


class DefiningContextType(str, Enum):
    """Types of defining context for categorical variation."""

    GENOMIC = "genomic"
    TRANSCRIPT = "transcript"
    PROTEIN = "protein"


class CanonicalVariationConstraint(GA4GHBase):
    """Constraint describing a class of canonical variations.

    Represents a constraint that defines a category of variation
    rather than a specific allele — e.g., any variant in a gene region.
    """

    type: str = "CanonicalVariationConstraint"
    defining_context: SequenceLocation | SequenceReference | None = None
    complement: bool = False
    copies: int | None = None
    copy_change: CopyChange | None = None


class CategoricalVariation(GA4GHBase):
    """A categorical variation: a class of related variants.

    Represents a set of variations grouped by a common
    biological or clinical meaning.
    """

    type: str = "CategoricalVariation"
    id: str | None = None
    label: str | None = None
    description: str | None = None
    members: list[CanonicalVariationConstraint] | None = None
    defining_context_type: DefiningContextType | None = None


class ProteinSequenceConsequence(GA4GHBase):
    """Consequence of a variation at the protein level."""

    type: str = "ProteinSequenceConsequence"
    defining_context: SequenceLocation | None = None
    label: str | None = None
    description: str | None = None



"""Tests for DRS client."""

from __future__ import annotations

import pytest
from pytest_httpx import HTTPXMock

from ai4gh.specs.drs.client import DRSClient
from ai4gh.specs.drs.models import DRSURI

DRS_OBJECT_RESPONSE = {
    "id": "drs-obj-1",
    "self_uri": "drs://example.com/drs-obj-1",
    "size": 1024,
    "created_time": "2024-01-01T00:00:00Z",
    "checksums": [{"checksum": "abc123", "type": "md5"}],
    "access_methods": [
        {"type": "https", "access_id": "https-1"},
    ],
}


@pytest.fixture
def drs_client() -> DRSClient:
    return DRSClient(base_url="https://example.com")


class TestDRSClient:
    def test_spec_attrs(self) -> None:
        client = DRSClient(base_url="https://example.com")
        assert client.spec_name == "drs"
        assert client.spec_version == "1.4.0"
        assert client.base_path == "/ga4gh/drs/v1"

    def test_url(self) -> None:
        client = DRSClient(base_url="https://example.com")
        assert client.url == "https://example.com/ga4gh/drs/v1"

    async def test_get_object(self, drs_client: DRSClient, httpx_mock: HTTPXMock) -> None:
        httpx_mock.add_response(
            url="https://example.com/ga4gh/drs/v1/objects/drs-obj-1",
            json=DRS_OBJECT_RESPONSE,
        )
        obj = await drs_client.get_object("drs-obj-1")
        assert obj.id == "drs-obj-1"
        assert obj.size == 1024
        assert len(obj.checksums) == 1

    async def test_get_object_expand(self, drs_client: DRSClient, httpx_mock: HTTPXMock) -> None:
        httpx_mock.add_response(json=DRS_OBJECT_RESPONSE)
        obj = await drs_client.get_object("drs-obj-1", expand=True)
        assert obj.id == "drs-obj-1"
        request = httpx_mock.get_requests()[0]
        assert "expand=true" in str(request.url)

    async def test_get_access_url(self, drs_client: DRSClient, httpx_mock: HTTPXMock) -> None:
        httpx_mock.add_response(
            url="https://example.com/ga4gh/drs/v1/objects/drs-obj-1/access/https-1",
            json={"url": "https://download.example.com/file.bam"},
        )
        access_url = await drs_client.get_access_url("drs-obj-1", "https-1")
        assert access_url.url == "https://download.example.com/file.bam"

    async def test_post_object(self, drs_client: DRSClient, httpx_mock: HTTPXMock) -> None:
        httpx_mock.add_response(json=DRS_OBJECT_RESPONSE)
        obj = await drs_client.post_object(
            "drs-obj-1",
            passports=["jwt-token-1"],
        )
        assert obj.id == "drs-obj-1"

    async def test_post_object_expand(self, drs_client: DRSClient, httpx_mock: HTTPXMock) -> None:
        httpx_mock.add_response(json=DRS_OBJECT_RESPONSE)
        obj = await drs_client.post_object("drs-obj-1", expand=True)
        assert obj.id == "drs-obj-1"

    def test_from_uri_string(self) -> None:
        client = DRSClient.from_uri("drs://example.com/obj-1")
        assert client.url == "https://example.com/ga4gh/drs/v1"

    def test_from_uri_object(self) -> None:
        uri = DRSURI("drs://example.com/obj-1")
        client = DRSClient.from_uri(uri)
        assert client.url == "https://example.com/ga4gh/drs/v1"



"""Pydantic models for GA4GH DRS v1.4."""

from __future__ import annotations

from datetime import datetime  # noqa: TC003 — needed at runtime by Pydantic
from enum import Enum
from urllib.parse import urlparse

from ai4gh.core.types import GA4GHBase


class AccessType(str, Enum):
    """DRS access method types."""

    s3 = "s3"
    gs = "gs"
    ftp = "ftp"
    gsiftp = "gsiftp"
    globus = "globus"
    htsget = "htsget"
    https = "https"
    file = "file"


class Checksum(GA4GHBase):
    """DRS checksum."""

    checksum: str
    type: str


class AccessURL(GA4GHBase):
    """DRS access URL."""

    url: str
    headers: list[str] | None = None


class DRSAccessMethod(GA4GHBase):
    """DRS access method."""

    type: AccessType
    access_url: AccessURL | None = None
    access_id: str | None = None
    cloud: str | None = None
    region: str | None = None
    available: bool | None = None


class ContentsObject(GA4GHBase):
    """DRS contents object (for bundles)."""

    name: str
    id: str | None = None
    drs_uri: list[str] | None = None
    contents: list[ContentsObject] | None = None


class DRSObject(GA4GHBase):
    """DRS object."""

    id: str
    self_uri: str
    size: int
    created_time: datetime
    updated_time: datetime | None = None
    version: str | None = None
    mime_type: str | None = None
    name: str | None = None
    checksums: list[Checksum]
    access_methods: list[DRSAccessMethod] | None = None
    contents: list[ContentsObject] | None = None
    description: str | None = None
    aliases: list[str] | None = None

    @property
    def is_bundle(self) -> bool:
        """Check if this DRS object is a bundle."""
        return self.contents is not None


class DRSURI:
    """Parser for DRS URIs (drs://host/id)."""

    def __init__(self, uri: str) -> None:
        if not uri.startswith("drs://"):
            raise ValueError(f"Invalid DRS URI: must start with drs:// (got {uri!r})")
        self._uri = uri
        parsed = urlparse(uri)
        self.host = parsed.hostname or ""
        self.id = parsed.path.lstrip("/")

    @property
    def base_url(self) -> str:
        """Convert to HTTPS base URL."""
        return f"https://{self.host}"

    def __str__(self) -> str:
        return self._uri

    def __repr__(self) -> str:
        return f"DRSURI({self._uri!r})"



"""Tests for DRS models."""

from __future__ import annotations

import pytest

from ai4gh.specs.drs.models import (
    DRSURI,
    AccessType,
    AccessURL,
    Checksum,
    ContentsObject,
    DRSAccessMethod,
    DRSObject,
)


class TestChecksum:
    def test_basic(self) -> None:
        c = Checksum(checksum="abc123", type="md5")
        assert c.checksum == "abc123"
        assert c.type == "md5"


class TestAccessURL:
    def test_basic(self) -> None:
        url = AccessURL(url="https://example.com/data")
        assert url.url == "https://example.com/data"
        assert url.headers is None

    def test_with_headers(self) -> None:
        url = AccessURL(url="https://example.com/data", headers=["Authorization: Bearer x"])
        assert url.headers == ["Authorization: Bearer x"]


class TestDRSAccessMethod:
    def test_basic(self) -> None:
        m = DRSAccessMethod(type=AccessType.https, access_id="abc")
        assert m.type == AccessType.https
        assert m.access_id == "abc"
        assert m.access_url is None

    def test_with_url(self) -> None:
        m = DRSAccessMethod(
            type=AccessType.s3,
            access_url=AccessURL(url="s3://bucket/key"),
        )
        assert m.access_url is not None
        assert m.access_url.url == "s3://bucket/key"


class TestContentsObject:
    def test_basic(self) -> None:
        c = ContentsObject(name="file.bam")
        assert c.name == "file.bam"
        assert c.id is None

    def test_nested(self) -> None:
        child = ContentsObject(name="child.bam", id="child-id")
        parent = ContentsObject(name="bundle", contents=[child])
        assert parent.contents is not None
        assert len(parent.contents) == 1
        assert parent.contents[0].name == "child.bam"


class TestDRSObject:
    def test_minimal(self) -> None:
        obj = DRSObject(
            id="drs-1",
            self_uri="drs://example.com/drs-1",
            size=1024,
            created_time="2024-01-01T00:00:00Z",
            checksums=[Checksum(checksum="abc", type="md5")],
        )
        assert obj.id == "drs-1"
        assert obj.size == 1024
        assert obj.is_bundle is False

    def test_bundle(self) -> None:
        obj = DRSObject(
            id="bundle-1",
            self_uri="drs://example.com/bundle-1",
            size=0,
            created_time="2024-01-01T00:00:00Z",
            checksums=[],
            contents=[ContentsObject(name="file.bam")],
        )
        assert obj.is_bundle is True

    def test_with_access_methods(self) -> None:
        obj = DRSObject(
            id="drs-2",
            self_uri="drs://example.com/drs-2",
            size=2048,
            created_time="2024-01-01T00:00:00Z",
            checksums=[Checksum(checksum="def", type="sha256")],
            access_methods=[
                DRSAccessMethod(type=AccessType.https, access_id="https-1"),
                DRSAccessMethod(type=AccessType.s3, access_id="s3-1"),
            ],
        )
        assert obj.access_methods is not None
        assert len(obj.access_methods) == 2


class TestDRSURI:
    def test_basic(self) -> None:
        uri = DRSURI("drs://example.com/abc123")
        assert uri.host == "example.com"
        assert uri.id == "abc123"
        assert uri.base_url == "https://example.com"

    def test_str(self) -> None:
        uri = DRSURI("drs://host.org/obj-id")
        assert str(uri) == "drs://host.org/obj-id"

    def test_repr(self) -> None:
        uri = DRSURI("drs://host.org/obj-id")
        assert "DRSURI" in repr(uri)

    def test_invalid_scheme(self) -> None:
        with pytest.raises(ValueError, match="must start with drs://"):
            DRSURI("https://example.com/abc")

    def test_nested_path(self) -> None:
        uri = DRSURI("drs://example.com/path/to/object")
        assert uri.id == "path/to/object"


class TestAccessType:
    def test_values(self) -> None:
        assert AccessType.s3.value == "s3"
        assert AccessType.https.value == "https"
        assert AccessType.gs.value == "gs"



"""Tests for ai4gh.core.auth."""

from __future__ import annotations

import pytest

from ai4gh.core.auth import (
    BearerTokenAuth,
    NoAuth,
    OAuth2Auth,
    PassportAuth,
    auto_detect_auth,
)
from ai4gh.core.config import AuthConfig
from ai4gh.core.exceptions import AuthError


class TestNoAuth:
    @pytest.mark.asyncio
    async def test_returns_empty_headers(self) -> None:
        auth = NoAuth()
        headers = await auth.get_headers()
        assert headers == {}


class TestBearerTokenAuth:
    @pytest.mark.asyncio
    async def test_returns_bearer_header(self) -> None:
        auth = BearerTokenAuth("my-token-123")
        headers = await auth.get_headers()
        assert headers == {"Authorization": "Bearer my-token-123"}


class TestOAuth2Auth:
    def test_init(self) -> None:
        auth = OAuth2Auth(
            token_url="https://auth.example.org/token",
            client_id="client-1",
            client_secret="secret-1",
        )
        assert auth._token_url == "https://auth.example.org/token"
        assert auth._access_token is None

    @pytest.mark.asyncio
    async def test_get_headers_fetches_token(self, httpx_mock) -> None:

        httpx_mock.add_response(
            url="https://auth.example.org/token",
            json={"access_token": "fetched-token-abc"},
        )
        auth = OAuth2Auth(
            token_url="https://auth.example.org/token",
            client_id="client-1",
            client_secret="secret-1",
        )
        headers = await auth.get_headers()
        assert headers == {"Authorization": "Bearer fetched-token-abc"}

    @pytest.mark.asyncio
    async def test_caches_token(self, httpx_mock) -> None:
        httpx_mock.add_response(
            url="https://auth.example.org/token",
            json={"access_token": "cached-token"},
        )
        auth = OAuth2Auth(
            token_url="https://auth.example.org/token",
            client_id="c",
            client_secret="s",
        )
        await auth.get_headers()
        # second call should not make another HTTP request
        headers = await auth.get_headers()
        assert headers == {"Authorization": "Bearer cached-token"}

    def test_invalidate(self) -> None:
        auth = OAuth2Auth(
            token_url="https://auth.example.org/token",
            client_id="c",
            client_secret="s",
        )
        auth._access_token = "old-token"
        auth.invalidate()
        assert auth._access_token is None

    @pytest.mark.asyncio
    async def test_fetch_failure_raises_auth_error(self, httpx_mock) -> None:
        httpx_mock.add_response(
            url="https://auth.example.org/token",
            status_code=401,
            text="unauthorized",
        )
        auth = OAuth2Auth(
            token_url="https://auth.example.org/token",
            client_id="c",
            client_secret="s",
        )
        with pytest.raises(AuthError, match="OAuth2 token fetch failed"):
            await auth.get_headers()


class TestPassportAuth:
    # Create a minimal JWT for testing (unsigned / no-verify)
    @pytest.fixture
    def sample_passport_token(self) -> str:
        """Create a minimal unsigned JWT for testing."""
        import base64
        import json

        header = base64.urlsafe_b64encode(
            json.dumps({"alg": "none", "typ": "JWT"}).encode()
        ).rstrip(b"=")
        payload = base64.urlsafe_b64encode(
            json.dumps(
                {
                    "sub": "user-123",
                    "ga4gh_passport_v1": [
                        {"type": "ControlledAccessGrants", "value": "dataset-1"},
                        {"type": "ResearcherStatus", "value": "approved"},
                    ],
                }
            ).encode()
        ).rstrip(b"=")
        return f"{header.decode()}.{payload.decode()}."

    @pytest.mark.asyncio
    async def test_get_headers(self, sample_passport_token: str) -> None:
        auth = PassportAuth(sample_passport_token, verify=False)
        headers = await auth.get_headers()
        assert headers["Authorization"].startswith("Bearer ")

    def test_decode_no_verify(self, sample_passport_token: str) -> None:
        auth = PassportAuth(sample_passport_token, verify=False)
        decoded = auth.decode(verify=False)
        assert decoded["sub"] == "user-123"

    def test_get_visas(self, sample_passport_token: str) -> None:
        auth = PassportAuth(sample_passport_token, verify=False)
        visas = auth.get_visas()
        assert len(visas) == 2
        assert visas[0]["type"] == "ControlledAccessGrants"

    def test_has_visa_type(self, sample_passport_token: str) -> None:
        auth = PassportAuth(sample_passport_token, verify=False)
        assert auth.has_visa_type("ControlledAccessGrants") is True
        assert auth.has_visa_type("LinkedIdentities") is False

    def test_decode_invalid_token(self) -> None:
        auth = PassportAuth("not-a-jwt", verify=False)
        with pytest.raises(AuthError, match="Invalid Passport token"):
            auth.decode(verify=True)


class TestAutoDetectAuth:
    def test_bearer_token(self) -> None:
        config = AuthConfig(token="my-token")
        auth = auto_detect_auth(config)
        assert isinstance(auth, BearerTokenAuth)

    def test_oauth2(self) -> None:
        config = AuthConfig(
            token_url="https://auth.example.org/token",
            client_id="client-1",
            client_secret="secret-1",
        )
        auth = auto_detect_auth(config)
        assert isinstance(auth, OAuth2Auth)

    def test_no_auth(self) -> None:
        config = AuthConfig()
        auth = auto_detect_auth(config)
        assert isinstance(auth, NoAuth)

    def test_token_takes_precedence(self) -> None:
        config = AuthConfig(
            token="my-token",
            token_url="https://auth.example.org/token",
            client_id="c",
            client_secret="s",
        )
        auth = auto_detect_auth(config)
        assert isinstance(auth, BearerTokenAuth)



"""Tests for TES models."""

from __future__ import annotations

from ai4gh.specs.tes.models import (
    TESCreateTaskResponse,
    TESExecutor,
    TESInput,
    TESListTasksResponse,
    TESOutput,
    TESResources,
    TESState,
    TESTask,
    TESTaskLog,
    TESView,
)


class TestTESState:
    def test_enum_values(self) -> None:
        assert TESState.UNKNOWN == "UNKNOWN"
        assert TESState.QUEUED == "QUEUED"
        assert TESState.RUNNING == "RUNNING"
        assert TESState.COMPLETE == "COMPLETE"
        assert TESState.EXECUTOR_ERROR == "EXECUTOR_ERROR"
        assert TESState.CANCELED == "CANCELED"
        assert TESState.PREEMPTED == "PREEMPTED"

    def test_from_string(self) -> None:
        assert TESState("RUNNING") is TESState.RUNNING


class TestTESView:
    def test_enum_values(self) -> None:
        assert TESView.MINIMAL == "MINIMAL"
        assert TESView.BASIC == "BASIC"
        assert TESView.FULL == "FULL"


class TestTESInput:
    def test_required_fields(self) -> None:
        inp = TESInput(path="/data/input.bam")
        assert inp.path == "/data/input.bam"
        assert inp.url is None
        assert inp.content is None

    def test_full(self) -> None:
        inp = TESInput(
            name="reads",
            description="Input BAM file",
            url="s3://bucket/input.bam",
            path="/data/input.bam",
            type="FILE",
            streamable=True,
        )
        assert inp.name == "reads"
        assert inp.url == "s3://bucket/input.bam"
        assert inp.streamable is True

    def test_inline_content(self) -> None:
        inp = TESInput(
            path="/data/script.sh",
            content="#!/bin/bash\necho hello",
        )
        assert inp.content.startswith("#!/bin/bash")


class TestTESOutput:
    def test_required_fields(self) -> None:
        out = TESOutput(url="s3://bucket/output.bam", path="/data/output.bam")
        assert out.url == "s3://bucket/output.bam"
        assert out.path == "/data/output.bam"

    def test_with_name(self) -> None:
        out = TESOutput(
            name="aligned",
            url="s3://bucket/aligned.bam",
            path="/data/aligned.bam",
            type="FILE",
        )
        assert out.name == "aligned"


class TestTESExecutor:
    def test_required_fields(self) -> None:
        ex = TESExecutor(image="ubuntu:20.04", command=["echo", "hello"])
        assert ex.image == "ubuntu:20.04"
        assert ex.command == ["echo", "hello"]
        assert ex.workdir is None

    def test_full(self) -> None:
        ex = TESExecutor(
            image="biocontainers/bwa:0.7.17",
            command=["bwa", "mem", "ref.fa", "reads.fq"],
            workdir="/work",
            stdin="/data/input.txt",
            stdout="/logs/stdout.log",
            stderr="/logs/stderr.log",
            env={"THREADS": "4"},
        )
        assert ex.workdir == "/work"
        assert ex.env == {"THREADS": "4"}


class TestTESResources:
    def test_defaults(self) -> None:
        res = TESResources()
        assert res.cpu_cores is None
        assert res.ram_gb is None
        assert res.preemptible is None

    def test_full(self) -> None:
        res = TESResources(
            cpu_cores=4,
            ram_gb=8.0,
            disk_gb=100.0,
            preemptible=True,
            zones=["us-east-1a"],
        )
        assert res.cpu_cores == 4
        assert res.ram_gb == 8.0
        assert res.preemptible is True


class TestTESTaskLog:
    def test_minimal(self) -> None:
        log = TESTaskLog()
        assert log.logs is None
        assert log.start_time is None

    def test_with_data(self) -> None:
        log = TESTaskLog(
            start_time="2024-01-01T00:00:00Z",
            end_time="2024-01-01T01:00:00Z",
            metadata={"host": "node-1"},
            system_logs=["Starting task"],
        )
        assert log.metadata == {"host": "node-1"}
        assert len(log.system_logs) == 1


class TestTESTask:
    def test_minimal(self) -> None:
        task = TESTask()
        assert task.id is None
        assert task.state is None
        assert task.executors is None

    def test_full_task(self) -> None:
        task = TESTask(
            id="task-1",
            state=TESState.COMPLETE,
            name="alignment",
            description="Align reads to reference",
            executors=[TESExecutor(image="bwa:latest", command=["bwa", "mem"])],
            inputs=[TESInput(path="/data/reads.fq", url="s3://bucket/reads.fq")],
            outputs=[TESOutput(url="s3://bucket/aligned.bam", path="/data/aligned.bam")],
            resources=TESResources(cpu_cores=4, ram_gb=8.0),
            tags={"project": "genomics"},
        )
        assert task.id == "task-1"
        assert task.state is TESState.COMPLETE
        assert len(task.executors) == 1
        assert task.tags == {"project": "genomics"}

    def test_from_dict(self) -> None:
        task = TESTask.model_validate(
            {
                "id": "task-2",
                "state": "RUNNING",
                "name": "variant_calling",
            }
        )
        assert task.state is TESState.RUNNING


class TestTESCreateTaskResponse:
    def test_creation(self) -> None:
        resp = TESCreateTaskResponse(id="task-new")
        assert resp.id == "task-new"


class TestTESListTasksResponse:
    def test_empty(self) -> None:
        resp = TESListTasksResponse()
        assert resp.tasks is None
        assert resp.next_page_token is None

    def test_with_tasks(self) -> None:
        resp = TESListTasksResponse(
            tasks=[TESTask(id="t1"), TESTask(id="t2")],
            next_page_token="page-2",
        )
        assert len(resp.tasks) == 2
        assert resp.next_page_token == "page-2"


class TestExtraFields:
    def test_allows_extra(self) -> None:
        task = TESTask.model_validate({"id": "task-1", "vendor_specific": "value"})
        assert task.vendor_specific == "value"



"""GA4GH refget v2 plugin."""

from __future__ import annotations

from ai4gh.specs.refget.client import RefgetClient
from ai4gh.specs.refget.models import RefgetAlias, RefgetMetadata

__all__ = [
    "RefgetAlias",
    "RefgetClient",
    "RefgetMetadata",
]



"""GA4GH Data Repository Service (DRS) v1.4 plugin."""

from __future__ import annotations

from ai4gh.specs.drs.client import DRSClient
from ai4gh.specs.drs.models import (
    DRSURI,
    AccessType,
    AccessURL,
    Checksum,
    ContentsObject,
    DRSAccessMethod,
    DRSObject,
)

__all__ = [
    "AccessType",
    "AccessURL",
    "Checksum",
    "ContentsObject",
    "DRSAccessMethod",
    "DRSClient",
    "DRSObject",
    "DRSURI",
]



"""Tests for MCP resources."""

from __future__ import annotations

import json

from ai4gh.mcp.resources import SPEC_DESCRIPTIONS, get_resource, list_resources


class TestListResources:
    def test_returns_list(self) -> None:
        resources = list_resources()
        assert isinstance(resources, list)

    def test_includes_specs_index(self) -> None:
        resources = list_resources()
        uris = [str(r.uri) for r in resources]
        assert "ga4gh://specs" in uris

    def test_includes_individual_specs(self) -> None:
        resources = list_resources()
        uris = [str(r.uri) for r in resources]
        for spec_id in SPEC_DESCRIPTIONS:
            assert f"ga4gh://specs/{spec_id}" in uris

    def test_total_count(self) -> None:
        resources = list_resources()
        # 1 index + N individual specs
        assert len(resources) == 1 + len(SPEC_DESCRIPTIONS)

    def test_resource_has_name_and_description(self) -> None:
        resources = list_resources()
        for r in resources:
            assert r.name
            assert r.description
            assert r.mimeType == "application/json"


class TestGetResource:
    def test_specs_index(self) -> None:
        result = json.loads(get_resource("ga4gh://specs"))
        assert "specifications" in result
        assert result["total"] == len(SPEC_DESCRIPTIONS)

    def test_individual_spec(self) -> None:
        result = json.loads(get_resource("ga4gh://specs/beacon"))
        assert result["name"] == "Beacon v2"
        assert result["version"] == "2.0.0"

    def test_individual_spec_drs(self) -> None:
        result = json.loads(get_resource("ga4gh://specs/drs"))
        assert result["name"] == "Data Repository Service (DRS)"

    def test_unknown_spec(self) -> None:
        result = json.loads(get_resource("ga4gh://specs/nonexistent"))
        assert "error" in result

    def test_unknown_uri(self) -> None:
        result = json.loads(get_resource("ga4gh://unknown"))
        assert "error" in result

    def test_all_specs_retrievable(self) -> None:
        for spec_id, spec_info in SPEC_DESCRIPTIONS.items():
            result = json.loads(get_resource(f"ga4gh://specs/{spec_id}"))
            assert result["name"] == spec_info["name"]
            assert result["version"] == spec_info["version"]



"""Pydantic models for GA4GH Data Connect v1."""

from __future__ import annotations

from typing import Any

from ai4gh.core.types import GA4GHBase


class ColumnSchema(GA4GHBase):
    """Schema for a single column in a Data Connect table."""

    name: str
    data_type: str | None = None
    description: str | None = None
    is_nullable: bool | None = None
    ontology_term: str | None = None


class TableSchema(GA4GHBase):
    """Schema for a Data Connect table."""

    columns: list[ColumnSchema] | None = None
    ref: str | None = None
    description: str | None = None


class TableInfo(GA4GHBase):
    """Information about a Data Connect table."""

    name: str
    description: str | None = None
    data_model: TableSchema | None = None


class TableListResponse(GA4GHBase):
    """Response for listing tables."""

    tables: list[TableInfo] | None = None
    pagination: dict[str, Any] | None = None


class TableData(GA4GHBase):
    """Data returned from a table query."""

    data: list[dict[str, Any]] | None = None
    data_model: TableSchema | None = None
    pagination: dict[str, Any] | None = None


class SearchRequest(GA4GHBase):
    """Data Connect search request."""

    query: str
    page_size: int | None = None
    page_token: str | None = None



"""Tests for htsget models."""

from __future__ import annotations

from ai4gh.specs.htsget.models import (
    HtsgetClass,
    HtsgetFormat,
    HtsgetTicket,
    HtsgetTicketData,
    HtsgetURL,
)


class TestHtsgetFormat:
    def test_values(self) -> None:
        assert HtsgetFormat.BAM.value == "BAM"
        assert HtsgetFormat.CRAM.value == "CRAM"
        assert HtsgetFormat.VCF.value == "VCF"
        assert HtsgetFormat.BCF.value == "BCF"


class TestHtsgetClass:
    def test_values(self) -> None:
        assert HtsgetClass.header.value == "header"
        assert HtsgetClass.body.value == "body"


class TestHtsgetURL:
    def test_basic(self) -> None:
        url = HtsgetURL(url="https://example.com/data/block1")
        assert url.url == "https://example.com/data/block1"
        assert url.headers is None
        assert url.data_class is None

    def test_with_headers(self) -> None:
        url = HtsgetURL(
            url="https://example.com/data/block1",
            headers={"Authorization": "Bearer token"},
        )
        assert url.headers is not None
        assert url.headers["Authorization"] == "Bearer token"

    def test_with_class(self) -> None:
        url = HtsgetURL(
            url="https://example.com/data/header",
            data_class=HtsgetClass.header,
        )
        assert url.data_class == HtsgetClass.header


class TestHtsgetTicketData:
    def test_basic(self) -> None:
        data = HtsgetTicketData(
            format=HtsgetFormat.BAM,
            urls=[HtsgetURL(url="https://example.com/block1")],
        )
        assert data.format == HtsgetFormat.BAM
        assert len(data.urls) == 1
        assert data.md5 is None

    def test_with_md5(self) -> None:
        data = HtsgetTicketData(
            format=HtsgetFormat.VCF,
            urls=[],
            md5="abc123",
        )
        assert data.md5 == "abc123"


class TestHtsgetTicket:
    def test_basic(self) -> None:
        ticket = HtsgetTicket(
            htsget=HtsgetTicketData(
                format=HtsgetFormat.CRAM,
                urls=[
                    HtsgetURL(url="https://example.com/header", data_class=HtsgetClass.header),
                    HtsgetURL(url="https://example.com/body", data_class=HtsgetClass.body),
                ],
            )
        )
        assert ticket.htsget.format == HtsgetFormat.CRAM
        assert len(ticket.htsget.urls) == 2



"""Passport validation and access decision utilities."""

from __future__ import annotations

import time
from typing import Any

import jwt

from ai4gh.core.exceptions import AuthError
from ai4gh.specs.passports.models import (
    AccessDecision,
    PassportClaims,
    VisaClaim,
    VisaType,
)


class PassportValidator:
    """Validates GA4GH Passport JWTs and makes access decisions.

    This provides higher-level validation than core.auth.PassportAuth,
    including structured Visa extraction, expiration checking, and
    access decision evaluation against required visa criteria.
    """

    def __init__(
        self,
        *,
        algorithms: list[str] | None = None,
    ) -> None:
        self._algorithms = algorithms or ["RS256", "ES256"]

    def decode(
        self,
        token: str,
        *,
        verify: bool = False,
        key: str = "",
    ) -> PassportClaims:
        """Decode a Passport JWT into structured claims.

        Args:
            token: The JWT string.
            verify: Whether to verify the signature.
            key: Public key for verification (required if verify=True).

        Returns:
            Structured passport claims.
        """
        options: dict[str, Any] = {"verify_signature": verify}
        if not verify:
            options["verify_exp"] = False
            options["verify_aud"] = False

        try:
            payload = jwt.decode(
                token,
                key=key,
                algorithms=self._algorithms,
                options=options,  # type: ignore[arg-type]
            )
        except jwt.InvalidTokenError as e:
            raise AuthError(f"Invalid Passport token: {e}") from e

        return PassportClaims.model_validate(payload)

    def extract_visas(
        self,
        claims: PassportClaims,
        *,
        visa_type: VisaType | None = None,
        source: str | None = None,
    ) -> list[VisaClaim]:
        """Extract visas from passport claims with optional filtering.

        Args:
            claims: Decoded passport claims.
            visa_type: Filter by visa type.
            source: Filter by visa source.

        Returns:
            List of matching visa claims.
        """
        visas = claims.ga4gh_passport_v1 or []
        if visa_type is not None:
            visas = [v for v in visas if v.type == visa_type]
        if source is not None:
            visas = [v for v in visas if v.source == source]
        return visas

    def is_expired(self, claims: PassportClaims) -> bool:
        """Check if the passport has expired.

        Args:
            claims: Decoded passport claims.

        Returns:
            True if expired, False otherwise.
        """
        if claims.exp is None:
            return False
        return time.time() > claims.exp

    def evaluate_access(
        self,
        claims: PassportClaims,
        *,
        required_type: VisaType,
        required_value: str | None = None,
        required_source: str | None = None,
    ) -> AccessDecision:
        """Evaluate whether a passport grants access based on criteria.

        Args:
            claims: Decoded passport claims.
            required_type: The visa type required for access.
            required_value: Required visa value pattern.
            required_source: Required visa source.

        Returns:
            An access decision with matching visas.
        """
        if self.is_expired(claims):
            return AccessDecision(
                granted=False,
                reason="Passport has expired",
            )

        matching = self.extract_visas(claims, visa_type=required_type, source=required_source)

        if required_value is not None:
            matching = [v for v in matching if v.value == required_value]

        if matching:
            return AccessDecision(
                granted=True,
                reason=f"Found {len(matching)} matching visa(s)",
                matching_visas=matching,
            )

        return AccessDecision(
            granted=False,
            reason=f"No visa of type {required_type.value} found",
        )



"""Tests for ai4gh.core.exceptions."""

from __future__ import annotations

from ai4gh.core.exceptions import (
    AI4GHError,
    AuthError,
    ConfigError,
    NotFoundError,
    PluginError,
    RateLimitError,
    ServerError,
    TransportError,
    ValidationError,
)


class TestExceptionHierarchy:
    def test_base_error(self) -> None:
        err = AI4GHError("test error")
        assert str(err) == "test error"
        assert err.status_code is None

    def test_base_error_with_status(self) -> None:
        err = AI4GHError("test", status_code=500)
        assert err.status_code == 500

    def test_config_error_inherits(self) -> None:
        err = ConfigError("bad config")
        assert isinstance(err, AI4GHError)

    def test_auth_error_inherits(self) -> None:
        err = AuthError("bad auth")
        assert isinstance(err, AI4GHError)

    def test_transport_error_inherits(self) -> None:
        err = TransportError("network error")
        assert isinstance(err, AI4GHError)

    def test_not_found_defaults(self) -> None:
        err = NotFoundError()
        assert err.status_code == 404
        assert isinstance(err, TransportError)

    def test_rate_limit_defaults(self) -> None:
        err = RateLimitError()
        assert err.status_code == 429
        assert err.retry_after is None

    def test_rate_limit_with_retry(self) -> None:
        err = RateLimitError(retry_after=30.0)
        assert err.retry_after == 30.0

    def test_server_error_inherits(self) -> None:
        err = ServerError("internal error", status_code=500)
        assert isinstance(err, TransportError)

    def test_plugin_error_inherits(self) -> None:
        err = PluginError("no plugin")
        assert isinstance(err, AI4GHError)

    def test_validation_error_inherits(self) -> None:
        err = ValidationError("invalid data")
        assert isinstance(err, AI4GHError)



"""Tests for TES client."""

from __future__ import annotations

import pytest
from pytest_httpx import HTTPXMock

from ai4gh.specs.tes.client import TESClient
from ai4gh.specs.tes.models import TESState, TESView


@pytest.fixture
def tes_client() -> TESClient:
    return TESClient(base_url="https://example.com")


class TestTESClient:
    def test_spec_attrs(self) -> None:
        client = TESClient(base_url="https://example.com")
        assert client.spec_name == "tes"
        assert client.spec_version == "1.1.0"
        assert client.base_path == "/ga4gh/tes/v1"

    def test_url(self) -> None:
        client = TESClient(base_url="https://example.com")
        assert client.url == "https://example.com/ga4gh/tes/v1"

    async def test_create_task(self, tes_client: TESClient, httpx_mock: HTTPXMock) -> None:
        httpx_mock.add_response(
            url="https://example.com/ga4gh/tes/v1/tasks",
            method="POST",
            json={"id": "task-new"},
        )
        task_def = {
            "name": "alignment",
            "executors": [{"image": "bwa:latest", "command": ["bwa", "mem"]}],
        }
        result = await tes_client.create_task(task_def)
        assert result.id == "task-new"

    async def test_get_task(self, tes_client: TESClient, httpx_mock: HTTPXMock) -> None:
        httpx_mock.add_response(
            url="https://example.com/ga4gh/tes/v1/tasks/task-1?view=MINIMAL",
            json={"id": "task-1", "state": "RUNNING"},
        )
        task = await tes_client.get_task("task-1")
        assert task.id == "task-1"
        assert task.state is TESState.RUNNING

    async def test_get_task_full_view(self, tes_client: TESClient, httpx_mock: HTTPXMock) -> None:
        httpx_mock.add_response(
            json={
                "id": "task-1",
                "state": "COMPLETE",
                "name": "alignment",
                "executors": [{"image": "bwa:latest", "command": ["bwa", "mem"]}],
            }
        )
        task = await tes_client.get_task("task-1", view=TESView.FULL)
        assert task.id == "task-1"
        request = httpx_mock.get_requests()[0]
        assert "view=FULL" in str(request.url)

    async def test_list_tasks(self, tes_client: TESClient, httpx_mock: HTTPXMock) -> None:
        httpx_mock.add_response(
            url="https://example.com/ga4gh/tes/v1/tasks?view=MINIMAL",
            json={
                "tasks": [
                    {"id": "t1", "state": "COMPLETE"},
                    {"id": "t2", "state": "RUNNING"},
                ],
                "next_page_token": "page2",
            },
        )
        resp = await tes_client.list_tasks()
        assert len(resp.tasks) == 2
        assert resp.next_page_token == "page2"

    async def test_list_tasks_with_filters(
        self, tes_client: TESClient, httpx_mock: HTTPXMock
    ) -> None:
        httpx_mock.add_response(json={"tasks": []})
        await tes_client.list_tasks(
            page_size=10,
            name_prefix="align",
            state="RUNNING",
            tag_key="project",
            tag_value="genomics",
        )
        request = httpx_mock.get_requests()[0]
        url = str(request.url)
        assert "page_size=10" in url
        assert "name_prefix=align" in url
        assert "state=RUNNING" in url
        assert "tag_key=project" in url

    async def test_cancel_task(self, tes_client: TESClient, httpx_mock: HTTPXMock) -> None:
        httpx_mock.add_response(
            url="https://example.com/ga4gh/tes/v1/tasks/task-1:cancel",
            method="POST",
            json={},
        )
        result = await tes_client.cancel_task("task-1")
        assert result == {}



"""GA4GH htsget v1.3 plugin."""

from __future__ import annotations

from ai4gh.specs.htsget.client import HtsgetClient
from ai4gh.specs.htsget.models import (
    HtsgetClass,
    HtsgetFormat,
    HtsgetTicket,
    HtsgetTicketData,
    HtsgetURL,
)

__all__ = [
    "HtsgetClass",
    "HtsgetClient",
    "HtsgetFormat",
    "HtsgetTicket",
    "HtsgetTicketData",
    "HtsgetURL",
]



"""Tests for CLI MCP commands."""

from __future__ import annotations

from typer.testing import CliRunner

from ai4gh.cli.main import app

runner = CliRunner()


class TestCLI:
    def test_version(self) -> None:
        result = runner.invoke(app, ["version"])
        assert result.exit_code == 0
        assert "ai4gh" in result.output

    def test_mcp_help(self) -> None:
        result = runner.invoke(app, ["mcp", "--help"])
        assert result.exit_code == 0
        assert "serve" in result.output

    def test_no_args_shows_help(self) -> None:
        result = runner.invoke(app, [])
        # no_args_is_help=True returns exit code 0 on typer >= 0.23
        assert result.exit_code in (0, 2)
        assert "mcp" in result.output or "version" in result.output



"""Tests for MCP tools."""

from __future__ import annotations

from ai4gh.mcp.tools import TOOLS, handle_tool_call, list_tools


class TestListTools:
    def test_returns_list(self) -> None:
        tools = list_tools()
        assert isinstance(tools, list)

    def test_has_tools(self) -> None:
        tools = list_tools()
        assert len(tools) == 13

    def test_tool_names(self) -> None:
        tools = list_tools()
        names = {t.name for t in tools}
        expected = {
            "drs_get_object",
            "beacon_query",
            "htsget_reads",
            "refget_sequence",
            "vrs_compute_id",
            "phenopacket_validate",
            "wes_submit_run",
            "tes_create_task",
            "trs_search_tools",
            "data_connect_search",
            "passport_decode",
            "duo_match",
            "pedigree_validate",
        }
        assert names == expected

    def test_all_tools_have_input_schema(self) -> None:
        for tool in TOOLS:
            assert tool.inputSchema
            assert tool.inputSchema["type"] == "object"
            assert "properties" in tool.inputSchema

    def test_all_tools_have_required_fields(self) -> None:
        for tool in TOOLS:
            assert "required" in tool.inputSchema

    def test_all_tools_have_description(self) -> None:
        for tool in TOOLS:
            assert tool.description


class TestHandleToolCall:
    async def test_unknown_tool(self) -> None:
        result = await handle_tool_call("nonexistent_tool", {})
        assert len(result) == 1
        assert "Unknown tool" in result[0].text

    async def test_phenopacket_validate_valid(self) -> None:
        result = await handle_tool_call(
            "phenopacket_validate",
            {"phenopacket": '{"id": "test-001", "subject": {"id": "patient-1"}}'},
        )
        assert len(result) == 1
        assert "Valid" in result[0].text or "test-001" in result[0].text

    async def test_phenopacket_validate_invalid_json(self) -> None:
        result = await handle_tool_call(
            "phenopacket_validate",
            {"phenopacket": "not-json"},
        )
        assert len(result) == 1
        assert "error" in result[0].text.lower() or "Error" in result[0].text

    async def test_vrs_compute_id(self) -> None:
        vrs_obj = '{"location": {"type": "SequenceLocation"}, "state": {"type": "LiteralSequenceExpression", "sequence": "T"}}'
        result = await handle_tool_call(
            "vrs_compute_id",
            {"type": "Allele", "definition": vrs_obj},
        )
        assert len(result) == 1
        assert "ga4gh:" in result[0].text

    async def test_passport_decode(self) -> None:
        import jwt

        token = jwt.encode(
            {
                "sub": "user-1",
                "iss": "https://issuer.example.com",
                "iat": 1700000000,
                "exp": 1800000000,
                "ga4gh_visa_v1": {
                    "type": "ControlledAccessGrants",
                    "value": "https://example.com/ds/1",
                    "source": "https://example.com",
                    "by": "dac",
                },
            },
            "secret",
            algorithm="HS256",
        )
        result = await handle_tool_call("passport_decode", {"token": token})
        assert len(result) == 1
        assert "user-1" in result[0].text

    async def test_pedigree_validate_valid(self) -> None:
        import json

        pedigree = json.dumps(
            {
                "individuals": [
                    {"id": "P1", "sex": "male", "affected_status": "affected"},
                    {"id": "P2", "sex": "female", "affected_status": "unaffected"},
                ],
                "relationships": [],
            }
        )
        result = await handle_tool_call(
            "pedigree_validate",
            {"pedigree": pedigree},
        )
        assert len(result) == 1
        assert "Valid" in result[0].text or "2 individuals" in result[0].text

    async def test_pedigree_validate_invalid_json(self) -> None:
        result = await handle_tool_call(
            "pedigree_validate",
            {"pedigree": "not-json"},
        )
        assert len(result) == 1
        assert "error" in result[0].text.lower() or "Error" in result[0].text

    async def test_duo_match(self) -> None:
        result = await handle_tool_call(
            "duo_match",
            {
                "conditions": [{"code": "DUO:0000042"}],  # GRU
                "purpose": {"codes": ["DUO:0000042"], "description": "General cancer research"},
            },
        )
        assert len(result) == 1



"""DRS v1.4 compliance test suite."""

from __future__ import annotations

import httpx

from ai4gh.compliance.runner import ComplianceSuite, TestResult, TestStatus, register_suite


@register_suite("drs")
class DRSComplianceSuite(ComplianceSuite):
    """Test GA4GH DRS API compliance."""

    spec_name = "drs"
    spec_version = "1.4.0"

    async def test_service_info(self) -> TestResult:
        """GET /ga4gh/drs/v1/service-info returns service info."""
        async with httpx.AsyncClient() as client:
            resp = await client.get(f"{self.base_url}/ga4gh/drs/v1/service-info")
        if resp.status_code != 200:
            return TestResult(
                name="test_service_info",
                status=TestStatus.FAIL,
                message=f"Expected 200, got {resp.status_code}",
                spec=self.spec_name,
                endpoint="/ga4gh/drs/v1/service-info",
            )
        data = resp.json()
        if "type" not in data:
            return TestResult(
                name="test_service_info",
                status=TestStatus.FAIL,
                message="Missing 'type' in service-info",
                spec=self.spec_name,
                endpoint="/ga4gh/drs/v1/service-info",
            )
        return TestResult(
            name="test_service_info",
            status=TestStatus.PASS,
            spec=self.spec_name,
            endpoint="/ga4gh/drs/v1/service-info",
        )

    async def test_get_object_not_found(self) -> TestResult:
        """GET /ga4gh/drs/v1/objects/<id> returns 404 for nonexistent object."""
        async with httpx.AsyncClient() as client:
            resp = await client.get(f"{self.base_url}/ga4gh/drs/v1/objects/nonexistent-id-12345")
        if resp.status_code == 404:
            return TestResult(
                name="test_get_object_not_found",
                status=TestStatus.PASS,
                spec=self.spec_name,
                endpoint="/ga4gh/drs/v1/objects/<id>",
            )
        if resp.status_code == 200:
            return TestResult(
                name="test_get_object_not_found",
                status=TestStatus.PASS,
                message="Server returned 200 (may have matching object)",
                spec=self.spec_name,
                endpoint="/ga4gh/drs/v1/objects/<id>",
            )
        return TestResult(
            name="test_get_object_not_found",
            status=TestStatus.FAIL,
            message=f"Expected 404, got {resp.status_code}",
            spec=self.spec_name,
            endpoint="/ga4gh/drs/v1/objects/<id>",
        )

    async def test_service_info_type_field(self) -> TestResult:
        """Service info type should indicate DRS."""
        async with httpx.AsyncClient() as client:
            resp = await client.get(f"{self.base_url}/ga4gh/drs/v1/service-info")
        if resp.status_code != 200:
            return TestResult(
                name="test_service_info_type_field",
                status=TestStatus.SKIP,
                message="Service info endpoint not available",
                spec=self.spec_name,
                endpoint="/ga4gh/drs/v1/service-info",
            )
        data = resp.json()
        stype = data.get("type", {})
        if isinstance(stype, dict) and stype.get("artifact") == "drs":
            return TestResult(
                name="test_service_info_type_field",
                status=TestStatus.PASS,
                spec=self.spec_name,
                endpoint="/ga4gh/drs/v1/service-info",
            )
        return TestResult(
            name="test_service_info_type_field",
            status=TestStatus.FAIL,
            message=f"Expected type.artifact='drs', got {stype}",
            spec=self.spec_name,
            endpoint="/ga4gh/drs/v1/service-info",
        )

    async def test_json_content_type(self) -> TestResult:
        """Service info should return JSON content type."""
        async with httpx.AsyncClient() as client:
            resp = await client.get(f"{self.base_url}/ga4gh/drs/v1/service-info")
        if resp.status_code != 200:
            return TestResult(
                name="test_json_content_type",
                status=TestStatus.SKIP,
                message="Service info not available",
                spec=self.spec_name,
                endpoint="/ga4gh/drs/v1/service-info",
            )
        ct = resp.headers.get("content-type", "")
        if "application/json" in ct:
            return TestResult(
                name="test_json_content_type",
                status=TestStatus.PASS,
                spec=self.spec_name,
                endpoint="/ga4gh/drs/v1/service-info",
            )
        return TestResult(
            name="test_json_content_type",
            status=TestStatus.FAIL,
            message=f"Expected application/json, got {ct}",
            spec=self.spec_name,
            endpoint="/ga4gh/drs/v1/service-info",
        )



"""GA4GH Pedigree plugin."""

from __future__ import annotations

from ai4gh.specs.pedigree.models import (
    AffectedStatus,
    KINRelationship,
    LifeStatus,
    Pedigree,
    PedigreeIndividual,
    PedigreeRelationship,
    Sex,
)

__all__ = [
    "AffectedStatus",
    "KINRelationship",
    "LifeStatus",
    "Pedigree",
    "PedigreeIndividual",
    "PedigreeRelationship",
    "Sex",
]



"""Tests for Data Connect models."""

from __future__ import annotations

from ai4gh.specs.data_connect.models import (
    ColumnSchema,
    SearchRequest,
    TableData,
    TableInfo,
    TableListResponse,
    TableSchema,
)


class TestColumnSchema:
    def test_required_fields(self) -> None:
        col = ColumnSchema(name="sample_id")
        assert col.name == "sample_id"
        assert col.data_type is None

    def test_full(self) -> None:
        col = ColumnSchema(
            name="age",
            data_type="integer",
            description="Patient age",
            is_nullable=True,
            ontology_term="EFO:0000246",
        )
        assert col.data_type == "integer"
        assert col.is_nullable is True
        assert col.ontology_term == "EFO:0000246"


class TestTableSchema:
    def test_minimal(self) -> None:
        schema = TableSchema()
        assert schema.columns is None

    def test_with_columns(self) -> None:
        schema = TableSchema(
            columns=[
                ColumnSchema(name="id", data_type="string"),
                ColumnSchema(name="age", data_type="integer"),
            ],
            description="Patient table schema",
        )
        assert len(schema.columns) == 2
        assert schema.description == "Patient table schema"


class TestTableInfo:
    def test_required_fields(self) -> None:
        info = TableInfo(name="patients")
        assert info.name == "patients"
        assert info.description is None
        assert info.data_model is None

    def test_with_schema(self) -> None:
        info = TableInfo(
            name="variants",
            description="Genomic variants",
            data_model=TableSchema(columns=[ColumnSchema(name="chrom", data_type="string")]),
        )
        assert info.data_model is not None
        assert len(info.data_model.columns) == 1


class TestTableListResponse:
    def test_empty(self) -> None:
        resp = TableListResponse()
        assert resp.tables is None
        assert resp.pagination is None

    def test_with_tables(self) -> None:
        resp = TableListResponse(
            tables=[
                TableInfo(name="patients"),
                TableInfo(name="variants"),
            ],
            pagination={"next_page_url": "https://example.com/tables?page=2"},
        )
        assert len(resp.tables) == 2
        assert "next_page_url" in resp.pagination


class TestTableData:
    def test_empty(self) -> None:
        data = TableData()
        assert data.data is None
        assert data.data_model is None

    def test_with_rows(self) -> None:
        data = TableData(
            data=[
                {"id": "1", "name": "Alice", "age": 30},
                {"id": "2", "name": "Bob", "age": 25},
            ],
            data_model=TableSchema(
                columns=[
                    ColumnSchema(name="id", data_type="string"),
                    ColumnSchema(name="name", data_type="string"),
                    ColumnSchema(name="age", data_type="integer"),
                ]
            ),
        )
        assert len(data.data) == 2
        assert data.data[0]["name"] == "Alice"
        assert len(data.data_model.columns) == 3

    def test_with_pagination(self) -> None:
        data = TableData(
            data=[{"id": "1"}],
            pagination={
                "next_page_url": "https://example.com/data?page=2",
            },
        )
        assert "next_page_url" in data.pagination


class TestSearchRequest:
    def test_required_fields(self) -> None:
        req = SearchRequest(query="SELECT * FROM patients")
        assert req.query == "SELECT * FROM patients"
        assert req.page_size is None

    def test_with_pagination(self) -> None:
        req = SearchRequest(
            query="SELECT * FROM variants WHERE chrom = '1'",
            page_size=100,
            page_token="token-1",
        )
        assert req.page_size == 100
        assert req.page_token == "token-1"


class TestExtraFields:
    def test_allows_extra(self) -> None:
        info = TableInfo.model_validate({"name": "table1", "custom_prop": "value"})
        assert info.custom_prop == "value"



"""GA4GH TRS (Tool Registry Service) v2.0.1 plugin."""

from __future__ import annotations

from ai4gh.specs.trs.client import TRSClient
from ai4gh.specs.trs.models import (
    DescriptorType,
    FileWrapper,
    ImageData,
    Tool,
    ToolClass,
    ToolVersion,
)

__all__ = [
    "DescriptorType",
    "FileWrapper",
    "ImageData",
    "TRSClient",
    "Tool",
    "ToolClass",
    "ToolVersion",
]



"""Tests for TRS models."""

from __future__ import annotations

from ai4gh.specs.trs.models import (
    DescriptorType,
    FileWrapper,
    ImageData,
    Tool,
    ToolClass,
    ToolSearchResponse,
    ToolVersion,
)


class TestDescriptorType:
    def test_enum_values(self) -> None:
        assert DescriptorType.CWL == "CWL"
        assert DescriptorType.WDL == "WDL"
        assert DescriptorType.NFL == "NFL"
        assert DescriptorType.GALAXY == "GALAXY"
        assert DescriptorType.SMK == "SMK"
        assert DescriptorType.JUPYTER == "JUPYTER"
        assert DescriptorType.OTHER == "OTHER"

    def test_from_string(self) -> None:
        assert DescriptorType("CWL") is DescriptorType.CWL


class TestToolClass:
    def test_minimal(self) -> None:
        tc = ToolClass()
        assert tc.id is None
        assert tc.name is None

    def test_full(self) -> None:
        tc = ToolClass(id="1", name="Workflow", description="A workflow")
        assert tc.id == "1"
        assert tc.name == "Workflow"


class TestFileWrapper:
    def test_with_content(self) -> None:
        fw = FileWrapper(content="class: Workflow\nsteps: []")
        assert fw.content.startswith("class:")
        assert fw.url is None

    def test_with_url(self) -> None:
        fw = FileWrapper(url="https://example.com/workflow.cwl")
        assert fw.url == "https://example.com/workflow.cwl"

    def test_with_checksum(self) -> None:
        fw = FileWrapper(
            content="data",
            checksum=[{"checksum": "abc123", "type": "sha256"}],
        )
        assert len(fw.checksum) == 1


class TestImageData:
    def test_minimal(self) -> None:
        img = ImageData()
        assert img.image_name is None
        assert img.size is None

    def test_full(self) -> None:
        img = ImageData(
            registry_host="docker.io",
            image_name="biocontainers/bwa",
            size=524288000,
            updated="2024-01-01T00:00:00Z",
            image_type="Docker",
        )
        assert img.registry_host == "docker.io"
        assert img.size == 524288000


class TestToolVersion:
    def test_minimal(self) -> None:
        tv = ToolVersion()
        assert tv.id is None
        assert tv.is_production is None

    def test_full(self) -> None:
        tv = ToolVersion(
            id="v1.0",
            name="Release 1.0",
            url="https://dockstore.org/tool/1/v1.0",
            is_production=True,
            descriptor_type=[DescriptorType.CWL, DescriptorType.WDL],
            author=["Alice", "Bob"],
            meta_version="2024-01-01",
            signed=False,
            images=[ImageData(image_name="tool:1.0", registry_host="docker.io")],
        )
        assert tv.is_production is True
        assert len(tv.descriptor_type) == 2
        assert tv.author == ["Alice", "Bob"]
        assert len(tv.images) == 1


class TestTool:
    def test_minimal(self) -> None:
        tool = Tool()
        assert tool.id is None
        assert tool.versions is None

    def test_full(self) -> None:
        tool = Tool(
            id="tool-1",
            url="https://dockstore.org/tool/1",
            name="BWA-MEM",
            organization="biocontainers",
            toolclass=ToolClass(id="1", name="Workflow"),
            description="BWA alignment workflow",
            meta_version="2024-01-01",
            has_checker=False,
            versions=[ToolVersion(id="v1.0", is_production=True)],
            aliases=["bwa", "bwa-mem"],
        )
        assert tool.id == "tool-1"
        assert tool.toolclass.name == "Workflow"
        assert len(tool.versions) == 1
        assert "bwa" in tool.aliases

    def test_from_dict(self) -> None:
        tool = Tool.model_validate(
            {
                "id": "tool-2",
                "name": "GATK",
                "organization": "Broad Institute",
            }
        )
        assert tool.name == "GATK"


class TestToolSearchResponse:
    def test_empty(self) -> None:
        resp = ToolSearchResponse()
        assert resp.tools is None

    def test_with_results(self) -> None:
        resp = ToolSearchResponse(
            tools=[Tool(id="t1"), Tool(id="t2")],
            next_page_token="next",
        )
        assert len(resp.tools) == 2
        assert resp.next_page_token == "next"


class TestExtraFields:
    def test_allows_extra(self) -> None:
        tool = Tool.model_validate({"id": "tool-1", "custom_registry": "dockstore"})
        assert tool.custom_registry == "dockstore"



"""Tests for TRS client."""

from __future__ import annotations

import pytest
from pytest_httpx import HTTPXMock

from ai4gh.specs.trs.client import TRSClient
from ai4gh.specs.trs.models import DescriptorType


@pytest.fixture
def trs_client() -> TRSClient:
    return TRSClient(base_url="https://example.com")


class TestTRSClient:
    def test_spec_attrs(self) -> None:
        client = TRSClient(base_url="https://example.com")
        assert client.spec_name == "trs"
        assert client.spec_version == "2.0.1"
        assert client.base_path == "/ga4gh/trs/v2"

    def test_url(self) -> None:
        client = TRSClient(base_url="https://example.com")
        assert client.url == "https://example.com/ga4gh/trs/v2"

    async def test_get_tool_classes(self, trs_client: TRSClient, httpx_mock: HTTPXMock) -> None:
        httpx_mock.add_response(
            url="https://example.com/ga4gh/trs/v2/toolClasses",
            json=[
                {"id": "1", "name": "Workflow"},
                {"id": "2", "name": "Tool"},
            ],
        )
        classes = await trs_client.get_tool_classes()
        assert len(classes) == 2
        assert classes[0].name == "Workflow"

    async def test_search_tools(self, trs_client: TRSClient, httpx_mock: HTTPXMock) -> None:
        httpx_mock.add_response(
            json=[
                {"id": "tool-1", "name": "BWA"},
                {"id": "tool-2", "name": "GATK"},
            ],
        )
        tools = await trs_client.search_tools(name="bwa")
        assert len(tools) == 2
        request = httpx_mock.get_requests()[0]
        assert "name=bwa" in str(request.url)

    async def test_search_tools_with_filters(
        self, trs_client: TRSClient, httpx_mock: HTTPXMock
    ) -> None:
        httpx_mock.add_response(json=[])
        await trs_client.search_tools(
            organization="broad",
            descriptor_type=DescriptorType.CWL,
            limit=10,
            offset=0,
        )
        request = httpx_mock.get_requests()[0]
        url = str(request.url)
        assert "organization=broad" in url
        assert "descriptorType=CWL" in url
        assert "limit=10" in url

    async def test_get_tool(self, trs_client: TRSClient, httpx_mock: HTTPXMock) -> None:
        httpx_mock.add_response(
            url="https://example.com/ga4gh/trs/v2/tools/tool-1",
            json={"id": "tool-1", "name": "BWA-MEM"},
        )
        tool = await trs_client.get_tool("tool-1")
        assert tool.id == "tool-1"
        assert tool.name == "BWA-MEM"

    async def test_get_tool_versions(self, trs_client: TRSClient, httpx_mock: HTTPXMock) -> None:
        httpx_mock.add_response(
            url="https://example.com/ga4gh/trs/v2/tools/tool-1/versions",
            json=[
                {"id": "v1.0", "is_production": True},
                {"id": "v2.0", "is_production": False},
            ],
        )
        versions = await trs_client.get_tool_versions("tool-1")
        assert len(versions) == 2
        assert versions[0].is_production is True

    async def test_get_tool_version(self, trs_client: TRSClient, httpx_mock: HTTPXMock) -> None:
        httpx_mock.add_response(
            url="https://example.com/ga4gh/trs/v2/tools/tool-1/versions/v1.0",
            json={"id": "v1.0", "name": "Release 1.0"},
        )
        version = await trs_client.get_tool_version("tool-1", "v1.0")
        assert version.id == "v1.0"

    async def test_get_descriptor(self, trs_client: TRSClient, httpx_mock: HTTPXMock) -> None:
        httpx_mock.add_response(
            url="https://example.com/ga4gh/trs/v2/tools/tool-1/versions/v1.0/CWL/descriptor",
            json={"content": "class: Workflow\nsteps: []"},
        )
        desc = await trs_client.get_descriptor("tool-1", "v1.0", DescriptorType.CWL)
        assert desc.content.startswith("class:")

    async def test_get_descriptor_files(self, trs_client: TRSClient, httpx_mock: HTTPXMock) -> None:
        httpx_mock.add_response(
            url="https://example.com/ga4gh/trs/v2/tools/tool-1/versions/v1.0/CWL/files",
            json=[
                {"content": "main.cwl", "url": "https://example.com/main.cwl"},
                {"content": "tools.cwl", "url": "https://example.com/tools.cwl"},
            ],
        )
        files = await trs_client.get_descriptor_files("tool-1", "v1.0", DescriptorType.CWL)
        assert len(files) == 2



"""GA4GH Passports v1.2 plugin."""

from __future__ import annotations

from ai4gh.specs.passports.models import (
    AccessDecision,
    PassportClaims,
    VisaAssertionSource,
    VisaClaim,
    VisaType,
)
from ai4gh.specs.passports.validator import PassportValidator

__all__ = [
    "AccessDecision",
    "PassportClaims",
    "PassportValidator",
    "VisaAssertionSource",
    "VisaClaim",
    "VisaType",
]



"""HTTP transport layer with retry, timeout, and rate limiting (httpx-based)."""

from __future__ import annotations

from typing import Any

import httpx
from tenacity import (
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential,
)

from ai4gh.core.config import Config
from ai4gh.core.exceptions import (
    NotFoundError,
    RateLimitError,
    ServerError,
    TransportError,
)


class Transport:
    """Async HTTP transport with retry, timeout, and rate limiting.

    Uses httpx for async HTTP/1.1 and HTTP/2 support.
    """

    def __init__(self, config: Config | None = None) -> None:
        self._config = config or Config()
        self._client: httpx.AsyncClient | None = None

    @property
    def client(self) -> httpx.AsyncClient:
        """Get or create the httpx async client."""
        if self._client is None or self._client.is_closed:
            self._client = httpx.AsyncClient(
                timeout=httpx.Timeout(self._config.timeout),
                verify=self._config.verify_ssl,
                follow_redirects=True,
            )
        return self._client

    async def close(self) -> None:
        """Close the HTTP client."""
        if self._client is not None and not self._client.is_closed:
            await self._client.aclose()
            self._client = None

    def _raise_for_status(self, response: httpx.Response) -> None:
        """Convert HTTP error responses to ai4gh exceptions."""
        if response.is_success:
            return

        status = response.status_code
        try:
            body = response.json()
            message = body.get("message", response.text)
        except Exception:
            message = response.text

        if status == 404:
            raise NotFoundError(message)
        if status == 429:
            retry_after = response.headers.get("Retry-After")
            raise RateLimitError(
                message,
                retry_after=float(retry_after) if retry_after else None,
            )
        if status >= 500:
            raise ServerError(message, status_code=status)
        raise TransportError(message, status_code=status)

    async def _request(
        self,
        method: str,
        url: str,
        *,
        headers: dict[str, str] | None = None,
        params: dict[str, Any] | None = None,
        json: Any = None,
        data: Any = None,
    ) -> httpx.Response:
        """Make an HTTP request with error handling."""
        try:
            response = await self.client.request(
                method,
                url,
                headers=headers,
                params=params,
                json=json,
                data=data,
            )
        except httpx.TimeoutException as e:
            raise TransportError(f"Request timed out: {e}") from e
        except httpx.ConnectError as e:
            raise TransportError(f"Connection failed: {e}") from e

        self._raise_for_status(response)
        return response

    @retry(
        retry=retry_if_exception_type((ServerError, RateLimitError)),
        stop=stop_after_attempt(3),
        wait=wait_exponential(multiplier=1, min=1, max=30),
        reraise=True,
    )
    async def get(
        self,
        url: str,
        *,
        headers: dict[str, str] | None = None,
        params: dict[str, Any] | None = None,
    ) -> httpx.Response:
        """HTTP GET with retry logic."""
        return await self._request("GET", url, headers=headers, params=params)

    @retry(
        retry=retry_if_exception_type((ServerError, RateLimitError)),
        stop=stop_after_attempt(3),
        wait=wait_exponential(multiplier=1, min=1, max=30),
        reraise=True,
    )
    async def post(
        self,
        url: str,
        *,
        headers: dict[str, str] | None = None,
        params: dict[str, Any] | None = None,
        json: Any = None,
        data: Any = None,
    ) -> httpx.Response:
        """HTTP POST with retry logic."""
        return await self._request(
            "POST", url, headers=headers, params=params, json=json, data=data
        )

    @retry(
        retry=retry_if_exception_type((ServerError, RateLimitError)),
        stop=stop_after_attempt(3),
        wait=wait_exponential(multiplier=1, min=1, max=30),
        reraise=True,
    )
    async def delete(
        self,
        url: str,
        *,
        headers: dict[str, str] | None = None,
    ) -> httpx.Response:
        """HTTP DELETE with retry logic."""
        return await self._request("DELETE", url, headers=headers)



"""Tests for ai4gh.core.registry."""

from __future__ import annotations

import pytest
from pytest_httpx import HTTPXMock

from ai4gh.core.registry import (
    ServiceInfoClient,
    ServiceRegistryClient,
    discover_services,
)

SERVICE_INFO_RESPONSE = {
    "id": "org.ga4gh.test",
    "name": "Test Service",
    "type": {
        "group": "org.ga4gh",
        "artifact": "beacon",
        "version": "2.0.0",
    },
    "organization": {
        "name": "Test Organization",
        "url": "https://example.org",
    },
    "version": "1.0.0",
}

SERVICES_RESPONSE = [
    {
        "id": "svc-beacon",
        "name": "Beacon",
        "type": {"group": "org.ga4gh", "artifact": "beacon", "version": "2.0.0"},
        "organization": {"name": "Test Org", "url": "https://example.org"},
    },
    {
        "id": "svc-drs",
        "name": "DRS",
        "type": {"group": "org.ga4gh", "artifact": "drs", "version": "1.4.0"},
        "organization": {"name": "Other Org", "url": "https://other.org"},
    },
]


class TestServiceInfoClient:
    @pytest.mark.asyncio
    async def test_get(self, httpx_mock: HTTPXMock) -> None:
        httpx_mock.add_response(url="https://example.org/service-info", json=SERVICE_INFO_RESPONSE)
        client = ServiceInfoClient("https://example.org")
        info = await client.get()
        assert info.id == "org.ga4gh.test"
        assert info.type.artifact == "beacon"

    @pytest.mark.asyncio
    async def test_strips_trailing_slash(self, httpx_mock: HTTPXMock) -> None:
        httpx_mock.add_response(url="https://example.org/service-info", json=SERVICE_INFO_RESPONSE)
        client = ServiceInfoClient("https://example.org/")
        info = await client.get()
        assert info.id == "org.ga4gh.test"


class TestServiceRegistryClient:
    @pytest.mark.asyncio
    async def test_list_services(self, httpx_mock: HTTPXMock) -> None:
        httpx_mock.add_response(url="https://registry.example.org/services", json=SERVICES_RESPONSE)
        client = ServiceRegistryClient("https://registry.example.org")
        services = await client.list_services()
        assert len(services) == 2
        assert services[0].id == "svc-beacon"

    @pytest.mark.asyncio
    async def test_get_service(self, httpx_mock: HTTPXMock) -> None:
        httpx_mock.add_response(
            url="https://registry.example.org/services/svc-beacon",
            json=SERVICES_RESPONSE[0],
        )
        client = ServiceRegistryClient("https://registry.example.org")
        svc = await client.get_service("svc-beacon")
        assert svc.id == "svc-beacon"

    @pytest.mark.asyncio
    async def test_discover_by_type(self, httpx_mock: HTTPXMock) -> None:
        httpx_mock.add_response(url="https://registry.example.org/services", json=SERVICES_RESPONSE)
        client = ServiceRegistryClient("https://registry.example.org")
        results = await client.discover(service_type="org.ga4gh:beacon:2.0.0")
        assert len(results) == 1
        assert results[0].id == "svc-beacon"

    @pytest.mark.asyncio
    async def test_discover_by_organization(self, httpx_mock: HTTPXMock) -> None:
        httpx_mock.add_response(url="https://registry.example.org/services", json=SERVICES_RESPONSE)
        client = ServiceRegistryClient("https://registry.example.org")
        results = await client.discover(organization="Other")
        assert len(results) == 1
        assert results[0].id == "svc-drs"

    @pytest.mark.asyncio
    async def test_discover_no_filters(self, httpx_mock: HTTPXMock) -> None:
        httpx_mock.add_response(url="https://registry.example.org/services", json=SERVICES_RESPONSE)
        client = ServiceRegistryClient("https://registry.example.org")
        results = await client.discover()
        assert len(results) == 2

    @pytest.mark.asyncio
    async def test_get_service_info(self, httpx_mock: HTTPXMock) -> None:
        httpx_mock.add_response(
            url="https://registry.example.org/service-info",
            json=SERVICE_INFO_RESPONSE,
        )
        client = ServiceRegistryClient("https://registry.example.org")
        info = await client.get_service_info()
        assert info.id == "org.ga4gh.test"


class TestDiscoverServices:
    @pytest.mark.asyncio
    async def test_multi_registry(self, httpx_mock: HTTPXMock) -> None:
        httpx_mock.add_response(
            url="https://r1.example.org/services",
            json=[SERVICES_RESPONSE[0]],
        )
        httpx_mock.add_response(
            url="https://r2.example.org/services",
            json=[SERVICES_RESPONSE[1]],
        )
        results = await discover_services(["https://r1.example.org", "https://r2.example.org"])
        assert len(results) == 2

    @pytest.mark.asyncio
    async def test_skips_failed_registries(self, httpx_mock: HTTPXMock) -> None:
        httpx_mock.add_response(
            url="https://good.example.org/services",
            json=[SERVICES_RESPONSE[0]],
        )
        # Transport retries ServerError up to 3 times, so mock all attempts
        for _ in range(3):
            httpx_mock.add_response(
                url="https://bad.example.org/services",
                status_code=500,
                text="error",
            )
        results = await discover_services(["https://good.example.org", "https://bad.example.org"])
        assert len(results) == 1
        assert results[0].id == "svc-beacon"

    @pytest.mark.asyncio
    async def test_with_type_filter(self, httpx_mock: HTTPXMock) -> None:
        httpx_mock.add_response(
            url="https://r1.example.org/services",
            json=SERVICES_RESPONSE,
        )
        results = await discover_services(
            ["https://r1.example.org"],
            service_type="org.ga4gh:drs:1.4.0",
        )
        assert len(results) == 1
        assert results[0].id == "svc-drs"



"""Pydantic models for GA4GH refget v2."""

from __future__ import annotations

from ai4gh.core.types import GA4GHBase


class RefgetMetadata(GA4GHBase):
    """Refget sequence metadata."""

    id: str
    md5: str | None = None
    trunc512: str | None = None
    ga4gh: str | None = None
    length: int
    aliases: list[RefgetAlias] | None = None


class RefgetAlias(GA4GHBase):
    """Refget sequence alias."""

    alias: str
    naming_authority: str


# Forward ref update
RefgetMetadata.model_rebuild()



"""Tests for Passports models."""

from __future__ import annotations

from ai4gh.specs.passports.models import (
    AccessDecision,
    PassportClaims,
    VisaAssertionSource,
    VisaClaim,
    VisaType,
)


class TestVisaType:
    def test_enum_values(self) -> None:
        assert VisaType.AFFILIATION_AND_ROLE == "AffiliationAndRole"
        assert VisaType.CONTROLLED_ACCESS_GRANTS == "ControlledAccessGrants"
        assert VisaType.LINKED_IDENTITIES == "LinkedIdentities"
        assert VisaType.RESEARCHER_STATUS == "ResearcherStatus"
        assert VisaType.ACCEPTED_TERMS_AND_POLICIES == "AcceptedTermsAndPolicies"


class TestVisaAssertionSource:
    def test_enum_values(self) -> None:
        assert VisaAssertionSource.SO == "so"
        assert VisaAssertionSource.SYSTEM == "system"


class TestVisaClaim:
    def test_minimal(self) -> None:
        visa = VisaClaim(
            type=VisaType.RESEARCHER_STATUS,
            value="https://example.com/researcher",
            source="https://idp.example.com",
        )
        assert visa.type is VisaType.RESEARCHER_STATUS
        assert visa.by is None

    def test_full(self) -> None:
        visa = VisaClaim(
            type=VisaType.CONTROLLED_ACCESS_GRANTS,
            value="https://dataset.example.com/ds001",
            source="https://dac.example.com",
            by=VisaAssertionSource.SO,
            asserted=1704067200,
            conditions=[[{"type": "AffiliationAndRole"}]],
        )
        assert visa.by is VisaAssertionSource.SO
        assert visa.asserted == 1704067200
        assert len(visa.conditions) == 1

    def test_from_dict(self) -> None:
        visa = VisaClaim.model_validate(
            {
                "type": "ResearcherStatus",
                "value": "https://example.com",
                "source": "https://idp.example.com",
            }
        )
        assert visa.type is VisaType.RESEARCHER_STATUS


class TestPassportClaims:
    def test_minimal(self) -> None:
        claims = PassportClaims()
        assert claims.iss is None
        assert claims.ga4gh_passport_v1 is None

    def test_full(self) -> None:
        claims = PassportClaims(
            iss="https://idp.example.com",
            sub="user@example.com",
            exp=1704153600,
            iat=1704067200,
            jti="unique-id",
            scope="openid ga4gh_passport_v1",
            ga4gh_passport_v1=[
                VisaClaim(
                    type=VisaType.RESEARCHER_STATUS,
                    value="https://example.com",
                    source="https://idp.example.com",
                )
            ],
        )
        assert claims.iss == "https://idp.example.com"
        assert len(claims.ga4gh_passport_v1) == 1


class TestAccessDecision:
    def test_granted(self) -> None:
        decision = AccessDecision(
            granted=True,
            reason="Found matching visa",
            matching_visas=[
                VisaClaim(
                    type=VisaType.CONTROLLED_ACCESS_GRANTS,
                    value="ds001",
                    source="https://dac.example.com",
                )
            ],
        )
        assert decision.granted is True
        assert len(decision.matching_visas) == 1

    def test_denied(self) -> None:
        decision = AccessDecision(
            granted=False,
            reason="No matching visa found",
        )
        assert decision.granted is False
        assert decision.matching_visas is None



"""Tests for Passport validator."""

from __future__ import annotations

import time

import jwt
import pytest

from ai4gh.core.exceptions import AuthError
from ai4gh.specs.passports.models import VisaType
from ai4gh.specs.passports.validator import PassportValidator


def _make_passport_jwt(
    visas: list[dict] | None = None,
    exp: int | None = None,
) -> str:
    """Create a test Passport JWT (unsigned, for testing only)."""
    payload: dict = {
        "iss": "https://idp.example.com",
        "sub": "user@example.com",
        "iat": int(time.time()),
    }
    if exp is not None:
        payload["exp"] = exp
    if visas is not None:
        payload["ga4gh_passport_v1"] = visas
    return jwt.encode(payload, "secret", algorithm="HS256")


@pytest.fixture
def validator() -> PassportValidator:
    return PassportValidator(algorithms=["HS256"])


class TestPassportValidator:
    def test_decode_basic(self, validator: PassportValidator) -> None:
        token = _make_passport_jwt()
        claims = validator.decode(token, verify=False)
        assert claims.iss == "https://idp.example.com"
        assert claims.sub == "user@example.com"

    def test_decode_with_visas(self, validator: PassportValidator) -> None:
        visas = [
            {
                "type": "ResearcherStatus",
                "value": "https://example.com",
                "source": "https://idp.example.com",
            }
        ]
        token = _make_passport_jwt(visas=visas)
        claims = validator.decode(token, verify=False)
        assert len(claims.ga4gh_passport_v1) == 1
        assert claims.ga4gh_passport_v1[0].type is VisaType.RESEARCHER_STATUS

    def test_decode_invalid_token(self, validator: PassportValidator) -> None:
        with pytest.raises(AuthError, match="Invalid Passport token"):
            validator.decode("not-a-jwt", verify=False)

    def test_extract_visas_all(self, validator: PassportValidator) -> None:
        visas = [
            {
                "type": "ResearcherStatus",
                "value": "v1",
                "source": "https://idp.example.com",
            },
            {
                "type": "ControlledAccessGrants",
                "value": "v2",
                "source": "https://dac.example.com",
            },
        ]
        token = _make_passport_jwt(visas=visas)
        claims = validator.decode(token, verify=False)
        extracted = validator.extract_visas(claims)
        assert len(extracted) == 2

    def test_extract_visas_by_type(self, validator: PassportValidator) -> None:
        visas = [
            {
                "type": "ResearcherStatus",
                "value": "v1",
                "source": "https://idp.example.com",
            },
            {
                "type": "ControlledAccessGrants",
                "value": "v2",
                "source": "https://dac.example.com",
            },
        ]
        token = _make_passport_jwt(visas=visas)
        claims = validator.decode(token, verify=False)
        extracted = validator.extract_visas(claims, visa_type=VisaType.CONTROLLED_ACCESS_GRANTS)
        assert len(extracted) == 1
        assert extracted[0].value == "v2"

    def test_extract_visas_by_source(self, validator: PassportValidator) -> None:
        visas = [
            {
                "type": "ResearcherStatus",
                "value": "v1",
                "source": "https://idp.example.com",
            },
            {
                "type": "ControlledAccessGrants",
                "value": "v2",
                "source": "https://dac.example.com",
            },
        ]
        token = _make_passport_jwt(visas=visas)
        claims = validator.decode(token, verify=False)
        extracted = validator.extract_visas(claims, source="https://dac.example.com")
        assert len(extracted) == 1

    def test_is_expired_false(self, validator: PassportValidator) -> None:
        token = _make_passport_jwt(exp=int(time.time()) + 3600)
        claims = validator.decode(token, verify=False)
        assert validator.is_expired(claims) is False

    def test_is_expired_true(self, validator: PassportValidator) -> None:
        token = _make_passport_jwt(exp=int(time.time()) - 3600)
        claims = validator.decode(token, verify=False)
        assert validator.is_expired(claims) is True

    def test_is_expired_no_exp(self, validator: PassportValidator) -> None:
        token = _make_passport_jwt()
        claims = validator.decode(token, verify=False)
        assert validator.is_expired(claims) is False

    def test_evaluate_access_granted(self, validator: PassportValidator) -> None:
        visas = [
            {
                "type": "ControlledAccessGrants",
                "value": "https://dataset.example.com/ds001",
                "source": "https://dac.example.com",
            }
        ]
        token = _make_passport_jwt(visas=visas, exp=int(time.time()) + 3600)
        claims = validator.decode(token, verify=False)
        decision = validator.evaluate_access(
            claims,
            required_type=VisaType.CONTROLLED_ACCESS_GRANTS,
            required_value="https://dataset.example.com/ds001",
        )
        assert decision.granted is True
        assert len(decision.matching_visas) == 1

    def test_evaluate_access_denied_no_visa(self, validator: PassportValidator) -> None:
        visas = [
            {
                "type": "ResearcherStatus",
                "value": "v1",
                "source": "https://idp.example.com",
            }
        ]
        token = _make_passport_jwt(visas=visas, exp=int(time.time()) + 3600)
        claims = validator.decode(token, verify=False)
        decision = validator.evaluate_access(
            claims,
            required_type=VisaType.CONTROLLED_ACCESS_GRANTS,
        )
        assert decision.granted is False

    def test_evaluate_access_denied_expired(self, validator: PassportValidator) -> None:
        visas = [
            {
                "type": "ControlledAccessGrants",
                "value": "ds001",
                "source": "https://dac.example.com",
            }
        ]
        token = _make_passport_jwt(visas=visas, exp=int(time.time()) - 3600)
        claims = validator.decode(token, verify=False)
        decision = validator.evaluate_access(
            claims,
            required_type=VisaType.CONTROLLED_ACCESS_GRANTS,
        )
        assert decision.granted is False
        assert "expired" in decision.reason



"""Tests for compliance test runner."""

from __future__ import annotations

from ai4gh.compliance.runner import (
    ComplianceSuite,
    SuiteResult,
    TestResult,
    TestStatus,
    get_suite,
    list_suites,
    register_suite,
)


class TestTestStatus:
    def test_values(self) -> None:
        assert TestStatus.PASS == "pass"
        assert TestStatus.FAIL == "fail"
        assert TestStatus.SKIP == "skip"
        assert TestStatus.ERROR == "error"


class TestTestResult:
    def test_defaults(self) -> None:
        r = TestResult(name="test_foo", status=TestStatus.PASS)
        assert r.name == "test_foo"
        assert r.status == TestStatus.PASS
        assert r.message == ""
        assert r.duration_ms == 0.0

    def test_with_message(self) -> None:
        r = TestResult(
            name="test_bar",
            status=TestStatus.FAIL,
            message="Expected 200",
            endpoint="/api/v1",
        )
        assert r.message == "Expected 200"
        assert r.endpoint == "/api/v1"


class TestSuiteResult:
    def test_empty(self) -> None:
        sr = SuiteResult(spec="test", base_url="http://example.com")
        assert sr.total == 0
        assert sr.passed == 0
        assert sr.failed == 0
        assert sr.skipped == 0
        assert sr.errors == 0
        assert sr.success_rate == 0.0

    def test_with_results(self) -> None:
        sr = SuiteResult(
            spec="test",
            base_url="http://example.com",
            results=[
                TestResult(name="t1", status=TestStatus.PASS),
                TestResult(name="t2", status=TestStatus.PASS),
                TestResult(name="t3", status=TestStatus.FAIL),
                TestResult(name="t4", status=TestStatus.SKIP),
                TestResult(name="t5", status=TestStatus.ERROR),
            ],
        )
        assert sr.total == 5
        assert sr.passed == 2
        assert sr.failed == 1
        assert sr.skipped == 1
        assert sr.errors == 1
        assert sr.success_rate == 40.0


class TestComplianceSuite:
    async def test_run_collects_methods(self) -> None:
        class MySuite(ComplianceSuite):
            spec_name = "test"

            def test_alpha(self) -> TestResult:
                return TestResult(name="test_alpha", status=TestStatus.PASS, spec="test")

            def test_beta(self) -> TestResult:
                return TestResult(name="test_beta", status=TestStatus.FAIL, spec="test")

        suite = MySuite(base_url="http://localhost")
        result = await suite.run()
        assert result.spec == "test"
        assert result.total == 2
        assert result.passed == 1
        assert result.failed == 1

    async def test_run_handles_exceptions(self) -> None:
        class BadSuite(ComplianceSuite):
            spec_name = "bad"

            def test_broken(self) -> TestResult:
                raise ValueError("boom")

        suite = BadSuite(base_url="http://localhost")
        result = await suite.run()
        assert result.total == 1
        assert result.errors == 1
        assert "boom" in result.results[0].message

    async def test_run_handles_none_return(self) -> None:
        class NullSuite(ComplianceSuite):
            spec_name = "null"

            def test_implicit_pass(self) -> None:
                pass

        suite = NullSuite(base_url="http://localhost")
        result = await suite.run()
        assert result.total == 1
        assert result.passed == 1

    async def test_run_async_methods(self) -> None:
        class AsyncSuite(ComplianceSuite):
            spec_name = "async"

            async def test_async_one(self) -> TestResult:
                return TestResult(name="test_async_one", status=TestStatus.PASS, spec="async")

        suite = AsyncSuite(base_url="http://localhost")
        result = await suite.run()
        assert result.total == 1
        assert result.passed == 1


class TestRegistry:
    def test_register_and_get(self) -> None:
        @register_suite("test_spec_xyz")
        class XYZSuite(ComplianceSuite):
            spec_name = "test_spec_xyz"

        assert get_suite("test_spec_xyz") is XYZSuite

    def test_get_nonexistent(self) -> None:
        assert get_suite("nonexistent_spec_abc") is None

    def test_list_suites(self) -> None:
        # Import suites to trigger registration
        import ai4gh.compliance.suites  # noqa: F401

        specs = list_suites()
        assert "beacon" in specs
        assert "drs" in specs
        assert "wes" in specs
        assert "tes" in specs
        assert "trs" in specs
        assert "htsget" in specs
        assert "refget" in specs



"""TES v1.1 API client."""

from __future__ import annotations

from typing import Any

from ai4gh.core.client import GA4GHClient
from ai4gh.specs.tes.models import (
    TESCreateTaskResponse,
    TESListTasksResponse,
    TESTask,
    TESView,
)


class TESClient(GA4GHClient):
    """Client for the GA4GH Task Execution Service (TES) v1.1.

    TES provides a standard API for submitting batch tasks with
    defined inputs, outputs, and executor commands.
    """

    spec_name = "tes"
    spec_version = "1.1.0"
    base_path = "/ga4gh/tes/v1"

    async def create_task(self, task: dict[str, Any]) -> TESCreateTaskResponse:
        """Create a new TES task.

        Args:
            task: The task definition dict.

        Returns:
            A response with the new task ID.
        """
        data = await self._request("POST", "/tasks", json=task)
        return TESCreateTaskResponse.model_validate(data)

    async def get_task(
        self,
        task_id: str,
        *,
        view: TESView = TESView.MINIMAL,
    ) -> TESTask:
        """Get a task by ID.

        Args:
            task_id: The task identifier.
            view: Detail level (MINIMAL, BASIC, FULL).

        Returns:
            The task with the requested level of detail.
        """
        params: dict[str, Any] = {"view": view.value}
        data = await self._request("GET", f"/tasks/{task_id}", params=params)
        return TESTask.model_validate(data)

    async def list_tasks(
        self,
        *,
        view: TESView = TESView.MINIMAL,
        page_size: int | None = None,
        page_token: str | None = None,
        name_prefix: str | None = None,
        state: str | None = None,
        tag_key: str | None = None,
        tag_value: str | None = None,
    ) -> TESListTasksResponse:
        """List tasks.

        Args:
            view: Detail level for listed tasks.
            page_size: Maximum number of tasks to return.
            page_token: Token for pagination.
            name_prefix: Filter by task name prefix.
            state: Filter by task state.
            tag_key: Filter by tag key.
            tag_value: Filter by tag value (requires tag_key).

        Returns:
            List of tasks with optional pagination token.
        """
        params: dict[str, Any] = {"view": view.value}
        if page_size is not None:
            params["page_size"] = page_size
        if page_token is not None:
            params["page_token"] = page_token
        if name_prefix is not None:
            params["name_prefix"] = name_prefix
        if state is not None:
            params["state"] = state
        if tag_key is not None:
            params["tag_key"] = tag_key
        if tag_value is not None:
            params["tag_value"] = tag_value
        data = await self._request("GET", "/tasks", params=params)
        return TESListTasksResponse.model_validate(data)

    async def cancel_task(self, task_id: str) -> dict[str, Any]:
        """Cancel a running task.

        Args:
            task_id: The task identifier.

        Returns:
            Empty dict on success.
        """
        data: dict[str, Any] = await self._request("POST", f"/tasks/{task_id}:cancel")
        return data



"""Tests for VRS v2 models."""

from __future__ import annotations

from ai4gh.specs.vrs.models import (
    Adjacency,
    Allele,
    CopyChange,
    CopyNumberChange,
    CopyNumberCount,
    Haplotype,
    LiteralSequenceExpression,
    ReferenceLengthExpression,
    SequenceLocation,
    SequenceReference,
    SequenceString,
    ga4gh_digest,
    ga4gh_identify,
)


class TestSequenceString:
    def test_basic(self) -> None:
        ss = SequenceString(sequence="ACGT")
        assert ss.sequence == "ACGT"
        assert ss.type == "SequenceString"


class TestSequenceReference:
    def test_basic(self) -> None:
        sr = SequenceReference(refgetAccession="SQ.abc123")
        assert sr.refgetAccession == "SQ.abc123"
        assert sr.type == "SequenceReference"

    def test_with_alphabet(self) -> None:
        sr = SequenceReference(
            refgetAccession="SQ.abc123",
            residueAlphabet="aa",
        )
        assert sr.residueAlphabet == "aa"


class TestSequenceLocation:
    def test_basic(self) -> None:
        loc = SequenceLocation(
            sequenceReference=SequenceReference(refgetAccession="SQ.abc"),
            start=100,
            end=200,
        )
        assert loc.start == 100
        assert loc.end == 200
        assert loc.type == "SequenceLocation"

    def test_with_range_coords(self) -> None:
        loc = SequenceLocation(
            sequenceReference=SequenceReference(refgetAccession="SQ.abc"),
            start=[10, 20],
            end=[30, 40],
        )
        assert loc.start == [10, 20]


class TestLiteralSequenceExpression:
    def test_basic(self) -> None:
        expr = LiteralSequenceExpression(sequence="T")
        assert expr.sequence == "T"
        assert expr.type == "LiteralSequenceExpression"


class TestReferenceLengthExpression:
    def test_basic(self) -> None:
        expr = ReferenceLengthExpression(length=5)
        assert expr.length == 5
        assert expr.type == "ReferenceLengthExpression"


class TestAllele:
    def test_basic(self) -> None:
        allele = Allele(
            location=SequenceLocation(
                sequenceReference=SequenceReference(refgetAccession="SQ.abc"),
                start=100,
                end=101,
            ),
            state=LiteralSequenceExpression(sequence="T"),
        )
        assert allele.type == "Allele"
        assert allele.location.start == 100
        assert allele.id is None

    def test_with_id(self) -> None:
        allele = Allele(
            id="ga4gh:VA.abc123",
            location=SequenceLocation(start=0, end=1),
            state=LiteralSequenceExpression(sequence="A"),
        )
        assert allele.id == "ga4gh:VA.abc123"


class TestHaplotype:
    def test_basic(self) -> None:
        allele1 = Allele(
            location=SequenceLocation(start=100, end=101),
            state=LiteralSequenceExpression(sequence="T"),
        )
        allele2 = Allele(
            location=SequenceLocation(start=200, end=201),
            state=LiteralSequenceExpression(sequence="G"),
        )
        hap = Haplotype(members=[allele1, allele2])
        assert hap.type == "Haplotype"
        assert len(hap.members) == 2


class TestCopyNumberCount:
    def test_basic(self) -> None:
        cnc = CopyNumberCount(
            location=SequenceLocation(
                sequenceReference=SequenceReference(refgetAccession="SQ.abc"),
                start=1000,
                end=2000,
            ),
            copies=3,
        )
        assert cnc.type == "CopyNumberCount"
        assert cnc.copies == 3

    def test_range_copies(self) -> None:
        cnc = CopyNumberCount(
            location=SequenceLocation(start=0, end=100),
            copies={"min": 2, "max": 4},
        )
        assert cnc.copies == {"min": 2, "max": 4}


class TestCopyNumberChange:
    def test_basic(self) -> None:
        cnc = CopyNumberChange(
            location=SequenceLocation(start=0, end=1000),
            copy_change=CopyChange.EFO_0030070,
        )
        assert cnc.type == "CopyNumberChange"
        assert cnc.copy_change == CopyChange.EFO_0030070


class TestAdjacency:
    def test_basic(self) -> None:
        adj = Adjacency(
            adj_left=SequenceLocation(start=100, end=100),
            adj_right=SequenceLocation(start=500, end=500),
        )
        assert adj.type == "Adjacency"
        assert adj.linker_sequence is None

    def test_with_linker(self) -> None:
        adj = Adjacency(
            adj_left=SequenceLocation(start=100, end=100),
            adj_right=SequenceLocation(start=500, end=500),
            linker_sequence="ACGT",
        )
        assert adj.linker_sequence == "ACGT"


class TestCopyChange:
    def test_gain(self) -> None:
        assert CopyChange.EFO_0030070.value == "efo:0030070"

    def test_loss(self) -> None:
        assert CopyChange.EFO_0030068.value == "efo:0030068"


class TestGA4GHDigest:
    def test_deterministic(self) -> None:
        data = {"type": "Allele", "location": {"start": 100}}
        d1 = ga4gh_digest(data)
        d2 = ga4gh_digest(data)
        assert d1 == d2

    def test_different_data(self) -> None:
        d1 = ga4gh_digest({"a": 1})
        d2 = ga4gh_digest({"a": 2})
        assert d1 != d2

    def test_key_order_invariant(self) -> None:
        d1 = ga4gh_digest({"b": 2, "a": 1})
        d2 = ga4gh_digest({"a": 1, "b": 2})
        assert d1 == d2


class TestGA4GHIdentify:
    def test_allele_format(self) -> None:
        data = {"type": "Allele", "location": {"start": 100}}
        identifier = ga4gh_identify(data, "VA")
        assert identifier.startswith("ga4gh:VA.")

    def test_haplotype_format(self) -> None:
        data = {"type": "Haplotype", "members": []}
        identifier = ga4gh_identify(data, "VH")
        assert identifier.startswith("ga4gh:VH.")

    def test_deterministic(self) -> None:
        data = {"type": "Allele", "state": "T"}
        id1 = ga4gh_identify(data, "VA")
        id2 = ga4gh_identify(data, "VA")
        assert id1 == id2



"""Pydantic models for GA4GH Pedigree.

Models for representing family pedigree structures using
standardized relationship types from the KIN ontology.
"""

from __future__ import annotations

from enum import Enum

from ai4gh.core.types import GA4GHBase


class Sex(str, Enum):
    """Biological sex."""

    MALE = "MALE"
    FEMALE = "FEMALE"
    OTHER = "OTHER_SEX"
    UNKNOWN = "UNKNOWN_SEX"


class AffectedStatus(str, Enum):
    """Disease affected status."""

    AFFECTED = "AFFECTED"
    UNAFFECTED = "UNAFFECTED"
    MISSING = "MISSING"


class LifeStatus(str, Enum):
    """Individual life status."""

    ALIVE = "ALIVE"
    DECEASED = "DECEASED"
    UNKNOWN = "UNKNOWN_STATUS"


class KINRelationship(str, Enum):
    """KIN ontology relationship types."""

    PARENT = "KIN:001"
    MOTHER = "KIN:002"
    FATHER = "KIN:003"
    SIBLING = "KIN:004"
    TWIN = "KIN:005"
    MONOZYGOTIC_TWIN = "KIN:006"
    DIZYGOTIC_TWIN = "KIN:007"
    CHILD = "KIN:008"
    HALF_SIBLING = "KIN:009"
    ADOPTED_CHILD = "KIN:010"
    FOSTER_CHILD = "KIN:011"
    GRANDPARENT = "KIN:012"
    GRANDCHILD = "KIN:013"
    SPOUSE = "KIN:014"
    COUSIN = "KIN:015"
    UNCLE_AUNT = "KIN:016"
    NEPHEW_NIECE = "KIN:017"


class PedigreeIndividual(GA4GHBase):
    """An individual in a pedigree."""

    individual_id: str
    sex: Sex | None = None
    affected_status: AffectedStatus | None = None
    life_status: LifeStatus | None = None
    maternal_id: str | None = None
    paternal_id: str | None = None
    date_of_birth: str | None = None


class PedigreeRelationship(GA4GHBase):
    """A relationship between two individuals."""

    subject_id: str
    object_id: str
    relationship: KINRelationship


class Pedigree(GA4GHBase):
    """A family pedigree."""

    id: str | None = None
    proband_id: str | None = None
    individuals: list[PedigreeIndividual]
    relationships: list[PedigreeRelationship] | None = None

    def get_individual(self, individual_id: str) -> PedigreeIndividual | None:
        """Find an individual by ID."""
        for ind in self.individuals:
            if ind.individual_id == individual_id:
                return ind
        return None

    def get_parents(self, individual_id: str) -> list[PedigreeIndividual]:
        """Get parents of an individual."""
        ind = self.get_individual(individual_id)
        if ind is None:
            return []
        parents = []
        if ind.maternal_id:
            mom = self.get_individual(ind.maternal_id)
            if mom:
                parents.append(mom)
        if ind.paternal_id:
            dad = self.get_individual(ind.paternal_id)
            if dad:
                parents.append(dad)
        return parents

    def get_siblings(self, individual_id: str) -> list[PedigreeIndividual]:
        """Get siblings of an individual (shared at least one parent)."""
        ind = self.get_individual(individual_id)
        if ind is None:
            return []
        siblings = []
        for other in self.individuals:
            if other.individual_id == individual_id:
                continue
            if (ind.maternal_id and other.maternal_id == ind.maternal_id) or (
                ind.paternal_id and other.paternal_id == ind.paternal_id
            ):
                siblings.append(other)
        return siblings

    def get_affected(self) -> list[PedigreeIndividual]:
        """Get all affected individuals."""
        return [ind for ind in self.individuals if ind.affected_status == AffectedStatus.AFFECTED]



"""Data Connect v1 API client."""

from __future__ import annotations

from typing import Any

from ai4gh.core.client import GA4GHClient
from ai4gh.specs.data_connect.models import TableData, TableInfo, TableListResponse


class DataConnectClient(GA4GHClient):
    """Client for the GA4GH Data Connect v1 API.

    Data Connect provides a standard API for searching tabular data
    using SQL-like expressions across federated datasets.
    """

    spec_name = "data_connect"
    spec_version = "1.0.0"
    base_path = ""

    async def list_tables(self) -> TableListResponse:
        """List available tables.

        Returns:
            A list of table information objects.
        """
        data = await self._request("GET", "/tables")
        return TableListResponse.model_validate(data)

    async def get_table_info(self, table_name: str) -> TableInfo:
        """Get information about a specific table.

        Args:
            table_name: The table name.

        Returns:
            Table information including schema.
        """
        data = await self._request("GET", f"/table/{table_name}/info")
        return TableInfo.model_validate(data)

    async def get_table_data(
        self,
        table_name: str,
        *,
        page_size: int | None = None,
        page_token: str | None = None,
    ) -> TableData:
        """Get data from a table.

        Args:
            table_name: The table name.
            page_size: Maximum number of rows to return.
            page_token: Token for pagination.

        Returns:
            Table data with optional pagination.
        """
        params: dict[str, Any] = {}
        if page_size is not None:
            params["page_size"] = page_size
        if page_token is not None:
            params["page_token"] = page_token
        data = await self._request("GET", f"/table/{table_name}/data", params=params or None)
        return TableData.model_validate(data)

    async def search(
        self,
        query: str,
        *,
        page_size: int | None = None,
        page_token: str | None = None,
    ) -> TableData:
        """Search using a SQL-like expression.

        Args:
            query: SQL-like query string.
            page_size: Maximum number of rows to return.
            page_token: Token for pagination.

        Returns:
            Query results with optional pagination.
        """
        body: dict[str, Any] = {"query": query}
        if page_size is not None:
            body["page_size"] = page_size
        if page_token is not None:
            body["page_token"] = page_token
        data = await self._request("POST", "/search", json=body)
        return TableData.model_validate(data)

    async def get_next_page(self, pagination_url: str) -> TableData:
        """Fetch the next page of results.

        Args:
            pagination_url: The next_page_url from pagination metadata.

        Returns:
            The next page of data.
        """
        auth_headers = await self._get_auth_headers()
        response = await self._transport.get(pagination_url, headers=auth_headers)
        return TableData.model_validate(response.json())



"""Compliance test suite for GA4GH specifications."""

from __future__ import annotations

from ai4gh.compliance.runner import (
    ComplianceSuite,
    SuiteResult,
    TestResult,
    TestStatus,
    get_suite,
    list_suites,
)

__all__ = [
    "ComplianceSuite",
    "SuiteResult",
    "TestResult",
    "TestStatus",
    "get_suite",
    "list_suites",
]
